# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Converts the AST to a logical query plan.

The plan does not try to be efficient or clever, at this point it is only trying to be correct.
"""

from enum import Enum
from enum import auto
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple

from orso.tools import random_string
from orso.types import OrsoTypes

from opteryx.exceptions import UnnamedColumnError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import format_expression
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import Node
from opteryx.planner.logical_planner import logical_planner_builders
from opteryx.planner.views import is_view
from opteryx.planner.views import view_as_plan
from opteryx.third_party.travers import Graph


class LogicalPlanStepType(int, Enum):
    Project = auto()  # field selection
    Filter = auto()  # tuple filtering
    Union = auto()  #  appending relations
    Explain = auto()  # EXPLAIN
    Difference = auto()  # relation interection
    Join = auto()  # all joins
    #    Containment = auto() # IN (maybe also EXISTS?)
    AggregateAndGroup = auto()  # group by
    Aggregate = auto()
    Scan = auto()  # read a dataset
    Show = auto()  # show a variable
    ShowColumns = auto()  # SHOW COLUMNS
    Set = auto()  # set a variable
    Limit = auto()  # limit and offset
    Order = auto()  # order by
    Distinct = auto()
    Exit = auto()
    HeapSort = auto()

    CTE = auto()
    Subquery = auto()
    FunctionDataset = auto()  # Unnest, GenerateSeries, values + Fake
    MetadataWriter = auto()


class LogicalPlan(Graph):
    pass


class LogicalPlanNode(Node):
    def copy(self) -> "Node":
        return LogicalPlanNode(**super().copy().properties)

    def __str__(self):
        try:
            # fmt:off
            node_type = self.node_type
            if node_type == LogicalPlanStepType.AggregateAndGroup:
                return f"AGGREGATE ({', '.join(format_expression(col) for col in self.aggregates)}) GROUP BY ({', '.join(format_expression(col) for col in self.groups)})"
            if node_type == LogicalPlanStepType.Aggregate:
                return f"AGGREGATE ({', '.join(format_expression(col) for col in self.aggregates)})"
            if node_type == LogicalPlanStepType.Distinct:
                distinct_on = ""
                if self.on is not None:
                    distinct_on = f" ON ({','.join(format_expression(col) for col in self.on)})"
                return f"DISTINCT{distinct_on}"
            if node_type == LogicalPlanStepType.Explain:
                return f"EXPLAIN{' ANALYZE' if self.analyze else ''}{(' (' + self.format + ')') if self.format else ''}"
            if node_type == LogicalPlanStepType.FunctionDataset:
                if self.function == "FAKE":
                    return f"FAKE ({', '.join(format_expression(arg) for arg in self.args)}{' AS ' + self.alias if self.alias else ''})"
                if self.function == "GENERATE_SERIES":
                    return f"GENERATE SERIES ({', '.join(format_expression(arg) for arg in self.args)}){' AS ' + self.alias if self.alias else ''}"
                if self.function == "VALUES":
                    return f"VALUES (({', '.join(self.columns)}) x {len(self.values)} AS {self.alias})"
                if self.function == "UNNEST":
                    return f"UNNEST ({', '.join(format_expression(arg) for arg in self.args)}{' AS ' + self.unnest_target if self.unnest_target else ''})"
                if self.function == "HTTP":
                    return f"HTTP ({self.url}) AS {self.alias}"
            if node_type == LogicalPlanStepType.Filter:
                return f"FILTER ({format_expression(self.condition)})"
            if node_type == LogicalPlanStepType.Join:
                distinct = " DISTINCT" if self.distinct else ""
                filters = f"({self.unnest_alias} IN ({', '.join(self.filters)}))" if self.filters else ""
                if self.on:
                    return f"{self.type.upper()} JOIN{distinct} ({format_expression(self.on, True)}){filters}"
                if self.using:
                    return f"{self.type.upper()} JOIN{distinct} (USING {','.join(map(format_expression, self.using))}){filters}"
                return f"{self.type.upper()}{distinct} {filters}"
            if node_type == LogicalPlanStepType.HeapSort:
                return f"HEAP SORT (LIMIT {self.limit}, ORDER BY {', '.join(format_expression(item[0]) + (' DESC' if item[1] =='descending' else '') for item in self.order_by)})"
            if node_type == LogicalPlanStepType.Limit:
                limit_str = f"LIMIT ({self.limit})" if self.limit is not None else ""
                offset_str = f" OFFSET ({self.offset})" if self.offset is not None else ""
                return (limit_str + offset_str).strip()
            if node_type == LogicalPlanStepType.Order:
                return f"ORDER BY ({', '.join(format_expression(item[0]) + (' DESC' if item[1] =='descending' else '') for item in self.order_by)})"
            if node_type == LogicalPlanStepType.Project:
                order_by_indicator = " +" if self.order_by_columns else ""
                return f"PROJECT ({', '.join(format_expression(col) for col in self.columns)}){order_by_indicator}"
            if node_type == LogicalPlanStepType.Scan:
                date_range = ""
                if self.start_date == self.end_date and self.start_date is not None:
                    date_range = f" FOR '{self.start_date}'"
                elif self.start_date is not None:
                    date_range = f" FOR '{self.start_date}' TO '{self.end_date}'"
                alias = ""
                if self.relation != self.alias:
                    alias = f" AS {self.alias}"
                columns = ""
                if self.columns:
                    columns = " [" + ", ".join(c.source_column for c in self.columns) + "]"
                predicates = ""
                if self.predicates:
                    predicates = " (" + " AND ".join(map(format_expression, self.predicates)) + ")"
                return f"READ ({self.relation}{alias}{date_range}{' WITH(' + ','.join(self.hints) + ')' if self.hints else ''}){columns}{predicates}"
            if node_type == LogicalPlanStepType.Set:
                return f"SET ({self.variable} TO {self.value.value})"
            if node_type == LogicalPlanStepType.Show:
                return f"SHOW ({', '.join(self.items)})"
            if node_type == LogicalPlanStepType.ShowColumns:
                return f"SHOW{' FULL' if self.full else ''}{' EXTENDED' if self.extended else ''} COLUMNS ({self.relation})"
            if node_type == LogicalPlanStepType.Subquery:
                return f"SUBQUERY{' AS ' + self.alias if self.alias else ''}"
            if node_type == LogicalPlanStepType.Union:
                columns = ""
                if self.columns:
                    columns = " [" + ", ".join(c.qualified_name for c in self.columns) + "]"
                return f"UNION {'' if self.modifier is None else self.modifier.upper()}{columns}"

            # fmt:on
        except Exception as err:
            import warnings

            warnings.warn(f"Problem drawing logical plan - {err}")
        return f"{str(self.node_type)[20:].upper()}"


def get_subplan_schemas(sub_plan: Graph) -> List[str]:
    """
    Retrieves schemas related to exit and entry points within a given sub-plan.

    This function iterates through functions named 'get_exit_points' and 'get_entry_points'
    in the `sub_plan` object to collect the schemas of the exit and entry points.

    Parameters:
        sub_plan: Graph
            The sub-plan object containing the necessary information for processing.

    Returns:
        List[str]:
            A list of schemas corresponding to the exit and entry points in the sub-plan.
    """
    aliases = set()

    def traverse(nid: dict):
        # get the actual node
        node = sub_plan[nid["name"]]
        # If this node is a subquery, capture its alias and stop traversing deeper
        if node.node_type == LogicalPlanStepType.Subquery:
            aliases.add(node.alias)
            return  # Stop traversing this branch as it's a separate scope
        if node.alias:
            aliases.add(node.alias)

        # Otherwise, continue traversing the children
        for child in nid.get("children", []):
            traverse(child)

    # Start the traversal from the provided node
    traverse(sub_plan.depth_first_search())

    return list(aliases)


"""
CLAUSE PLANNERS
"""


def extract_ctes(branch, planner):
    ctes = {}
    if branch.get("Query", branch).get("with"):
        for _ast in branch.get("Query", branch)["with"]["cte_tables"]:
            alias = _ast.get("alias")["name"]["value"]
            ctes[alias] = planner(_ast["query"]["body"])
    return ctes


def extract_value(clause):
    if len(clause) == 1:
        return logical_planner_builders.build(clause[0])
    return [logical_planner_builders.build(token) for token in clause]


def extract_variable(clause):
    if len(clause) == 1:
        return clause[0]["value"]
    return [token["value"] for token in clause]


def extract_simple_filter(filters, identifier: str = "Name"):
    if "Like" in filters:
        left = Node(NodeType.IDENTIFIER, value=identifier)
        right = Node(NodeType.LITERAL, type=OrsoTypes.VARCHAR, value=filters["Like"])
        root = Node(
            NodeType.COMPARISON_OPERATOR,
            value="ILike",  # we're case insensitive for SHOW filters
            left=left,
            right=right,
        )
        return root
    if "Where" in filters:
        root = logical_planner_builders.build(filters["Where"])
        return root


def _table_name(branch):
    keys = ("Table", "Derived")
    for key in keys:
        if key in branch["relation"]:
            break
    if branch["relation"][key]["alias"]:
        return branch["relation"][key]["alias"]["name"]["value"]
    return ".".join(part["value"] for part in branch["relation"][key]["name"])


def inner_query_planner(ast_branch):
    if "Query" in ast_branch:
        # Sometimes we get a full query plan here (e.g. when queries in set
        # functions are in parenthesis)
        return plan_query(ast_branch)

    inner_plan = LogicalPlan()
    step_id = None

    # TOP used?
    if ast_branch["Select"].get("top") is not None:
        raise UnsupportedSyntaxError(
            "SELECT TOP to limit number of returned records not supported, use LIMIT instead."
        )

    # from
    _relations = ast_branch["Select"].get("from", [])
    for relation in _relations:
        step_id, sub_plan = create_node_relation(relation)
        inner_plan += sub_plan

    # If there's any peer relations, they are implicit cross joins
    if len(_relations) > 2:
        raise UnsupportedSyntaxError("Cannot CROSS JOIN more than two relations.")
    if len(_relations) > 1:
        join_step = LogicalPlanNode(node_type=LogicalPlanStepType.Join)
        join_step.type = "cross join"

        join_step.right_relation_names = [_table_name(_relations[0])]
        join_step.left_relation_names = [_table_name(_relations[1])]

        step_id = random_string()
        inner_plan.add_node(step_id, join_step)
        for relation in _relations:
            inner_plan.add_edge(relation["step_id"], step_id)

    # If there's no relations, use $no_table
    if len(_relations) == 0:
        step_id, sub_plan = create_node_relation(
            {
                "relation": {
                    "Table": {
                        "name": [{"value": "$no_table"}],
                        "args": None,
                        "alias": None,
                        "with_hints": [],
                    }
                }
            }
        )
        inner_plan += sub_plan

    # selection
    _selection = logical_planner_builders.build(ast_branch["Select"].get("selection"))
    if _selection:
        if len(_relations) == 0:
            raise UnsupportedSyntaxError("Statement has a WHERE clause but no FROM clause.")
        selection_step = LogicalPlanNode(node_type=LogicalPlanStepType.Filter)
        selection_step.condition = _selection
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, selection_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

    # groups
    _projection = logical_planner_builders.build(ast_branch["Select"].get("projection")) or []
    if len(_projection) > 1 and any(p.node_type == NodeType.WILDCARD for p in _projection):
        from opteryx.exceptions import SqlError

        raise SqlError("SELECT * cannot coexist with additional columns.")
    _aggregates = get_all_nodes_of_type(_projection, select_nodes=(NodeType.AGGREGATOR,))
    _groups = logical_planner_builders.build(ast_branch["Select"].get("group_by"))[0]
    if _groups is not None and _groups != []:
        if any(p.node_type == NodeType.WILDCARD for p in _projection):
            raise UnsupportedSyntaxError(
                "SELECT * cannot be used with GROUP BY, fields in the SELECT must be aggregates or in the GROUP BY clause."
            )

        group_step = LogicalPlanNode(node_type=LogicalPlanStepType.AggregateAndGroup)
        group_step.groups = _groups
        group_step.aggregates = _aggregates
        group_step.projection = _projection
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, group_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)
    # aggregates
    elif len(_aggregates) > 0:
        aggregate_step = LogicalPlanNode(node_type=LogicalPlanStepType.Aggregate)
        aggregate_step.groups = _groups
        aggregate_step.aggregates = _aggregates

        known_columns = set(get_all_nodes_of_type(_groups + _aggregates, (NodeType.IDENTIFIER,)))
        project_columns = set(get_all_nodes_of_type(_projection, (NodeType.IDENTIFIER,)))

        if project_columns - known_columns:
            from opteryx.exceptions import SqlError

            column = (project_columns - known_columns).pop().source_column
            error = f"Column '{column}' must appear in the `GROUP BY` clause or must be part of an aggregate function. Either add it to the `GROUP BY` list, or add an aggregation such as `MIN({column})`."
            raise SqlError(error)

        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, aggregate_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

    # pre-process part of the order by before the projection
    _order_by = ast_branch.get("order_by")
    _order_by_columns_not_in_projection = []
    _order_by_columns = []
    if _order_by and _order_by.get("exprs"):
        _order_by = [
            (
                logical_planner_builders.build(item["expr"]),
                True if item["asc"] is None else item["asc"],
            )
            for item in _order_by["exprs"]
        ]
        if any(c[0].node_type == NodeType.LITERAL for c in _order_by):
            raise UnsupportedSyntaxError("Cannot ORDER BY constant values")
        _order_by_columns = [exp[0] for exp in _order_by]

    # projection
    if not (
        len(_projection) == 1
        and _projection[0].node_type == NodeType.WILDCARD
        and _projection[0].value is None
    ):
        for column in _projection:
            if column.node_type == NodeType.LITERAL and column.type == OrsoTypes.ARRAY:
                if ast_branch["Select"].get("distinct"):
                    raise UnsupportedSyntaxError(
                        "Values cannot be parenthesised in the SELECT clause. Did you mean DISTINCT ON(cols) cols FROM ?"
                    )
                raise UnsupportedSyntaxError("Values cannot be parenthesised in the SELECT clause.")

        # ORDER BY needing to be able to order by columns not in the projection
        # whilst being able to order by aliases created by the projection means
        # we need to do specific checks
        if _order_by_columns:
            # Collect qualified names and aliases from projection columns
            projection_qualified_names = {
                proj_col.qualified_name for proj_col in _projection if proj_col.qualified_name
            }.union({f".{proj_col.alias}" for proj_col in _projection if proj_col.alias})

            # Collect expression columns from projection
            projection_expressions = {
                format_expression(proj_col)
                for proj_col in _projection
                if proj_col.node_type != NodeType.IDENTIFIER
            }

            # Remove columns from ORDER BY that are directly in the projection, aliased, or have the same expression
            _order_by_columns_not_in_projection = [
                ord_col
                for ord_col in _order_by_columns
                if ord_col.qualified_name not in projection_qualified_names
                and f".{ord_col.source_column}" not in projection_qualified_names
                and ord_col.qualified_name
                not in [f".{proj_col.source_column}" for proj_col in _projection]
                and format_expression(ord_col) not in projection_expressions
            ]

            # Remove columns from ORDER BY that match the source of a wildcard in the projection
            for proj_col in [pc for pc in _projection if pc.node_type == NodeType.WILDCARD]:
                _order_by_columns_not_in_projection = [
                    ord_col
                    for ord_col in _order_by_columns_not_in_projection
                    if ord_col.source != proj_col.value[0]
                ]

        project_step = LogicalPlanNode(node_type=LogicalPlanStepType.Project)
        project_step.columns = _projection
        project_step.order_by_columns = _order_by_columns_not_in_projection
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, project_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)
    else:
        _order_by_columns_not_in_projection = []

    # having
    _having = logical_planner_builders.build(ast_branch["Select"].get("having"))
    if _having:
        having_step = LogicalPlanNode(node_type=LogicalPlanStepType.Filter)
        having_step.condition = _having
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, having_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

    # distinct
    if ast_branch["Select"].get("distinct"):
        distinct_step = LogicalPlanNode(node_type=LogicalPlanStepType.Distinct)
        if isinstance(ast_branch["Select"]["distinct"], dict):
            distinct_step.on = logical_planner_builders.build(
                ast_branch["Select"]["distinct"]["On"]
            )
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, distinct_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

        if distinct_step.on and _projection[0].source_column == "FROM":
            cols = ", ".join([format_expression(c) for c in distinct_step.on])
            raise UnsupportedSyntaxError(
                f"Did you mean 'SELECT DISTINCT ON ({cols}) {cols} FROM {_projection[0].alias};'?"
            )

    # order
    if _order_by:
        order_step = LogicalPlanNode(node_type=LogicalPlanStepType.Order)
        order_step.order_by = _order_by
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, order_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

    # limit/offset
    _limit = ast_branch.get("limit")
    _offset = ast_branch.get("offset")
    if _limit or _offset:
        limit_step = LogicalPlanNode(node_type=LogicalPlanStepType.Limit)
        limit_step.limit = None if _limit is None else logical_planner_builders.build(_limit).value
        limit_step.offset = (
            None if _offset is None else logical_planner_builders.build(_offset).value
        )
        previous_step_id, step_id = step_id, random_string()
        inner_plan.add_node(step_id, limit_step)
        if previous_step_id is not None:
            inner_plan.add_edge(previous_step_id, step_id)

    # add the exit node
    exit_node = LogicalPlanNode(node_type=LogicalPlanStepType.Exit)
    exit_node.columns = _projection
    previous_step_id, step_id = step_id, random_string()
    inner_plan.add_node(step_id, exit_node)
    if previous_step_id is not None:
        inner_plan.add_edge(previous_step_id, step_id)

    return inner_plan


"""
STATEMENT PLANNERS
"""


def process_join_tree(join: dict) -> LogicalPlanNode:
    """
    Processes a join tree from the AST and returns a LogicalPlanNode representing the join.
    """

    def extract_join_type(join: dict) -> str:
        """
        Extracts the type of the join from the AST node representing the join.
        """
        join_operator = join["join_operator"]

        if join_operator == {"Inner": "Natural"}:
            return "natural join"
        elif join_operator == "CrossJoin":
            return "cross join"

        join_operator = next(iter(join["join_operator"]))

        return {
            "FullOuter": "full outer",
            "Inner": "inner",
            "LeftAnti": "left anti",
            "LeftOuter": "left outer",
            "LeftSemi": "left semi",
            "RightAnti": "right anti",
            "RightOuter": "right outer",
            "RightSemi": "right semi",
            "CrossJoin": "cross join",  # should never match, here for completeness
            "Natural": "natural join",  # should never match, here for completeness
        }.get(join_operator)

    def extract_join_condition(join: dict) -> Tuple[Optional[str], Optional[List[str]]]:
        """
        Extracts the join's limiting condition from the AST node representing the join.
        """
        join_operator = join["join_operator"]
        if not isinstance(join_operator, dict):
            return None, None

        join_on = None
        join_using = None

        join_operator = next(iter(join_operator))
        join_condition = next(iter(join["join_operator"][join_operator]))
        if join_condition == "On":
            join_on = logical_planner_builders.build(
                join["join_operator"][join_operator][join_condition]
            )
        if join_condition == "Using":
            join_using = [
                logical_planner_builders.build({"Identifier": identifier})
                for identifier in join["join_operator"][join_operator][join_condition]
            ]

        return join_on, join_using

    def extract_unnest_dataset(join: dict, join_type: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Extracts information for an UNNEST dataset from the AST node representing the join.
        """
        if (
            join["relation"].get("Table", {}).get("name", [{}])[0].get("value", "").upper()
            != "UNNEST"
        ):
            return None, None

        if join_type not in ("cross join", "inner"):
            raise UnsupportedSyntaxError("JOIN on UNNEST only supported for CROSS and INNER joins.")
        unnest_column = logical_planner_builders.build(join["relation"]["Table"]["args"]["args"][0])
        if join["relation"]["Table"].get("alias") is None:
            raise UnnamedColumnError(
                "Column created by UNNEST has no name, use AS to name the column."
            )
        unnest_alias = join["relation"]["Table"]["alias"]["name"]["value"]

        # return the column we're going to UNNEST and what the new column is called
        return unnest_column, unnest_alias

    join_step = LogicalPlanNode(node_type=LogicalPlanStepType.Join)

    join_step.type = extract_join_type(join)
    join_step.on, join_step.using = extract_join_condition(join)
    # At this stage, CROSS JOIN UNNEST are represented in a single JOIN node
    join_step.unnest_column, join_step.unnest_alias = extract_unnest_dataset(join, join_step.type)

    return join_step


def create_node_relation(relation):
    sub_plan = LogicalPlan()
    root_node = None

    if "Derived" in relation["relation"]:
        if relation["relation"]["Derived"]["subquery"]:
            subquery = relation["relation"]["Derived"]
            if "Values" not in subquery["subquery"]["body"]:
                # SUBQUERY nodes wrap other queries and the result is available as a relation in
                # the parent query.
                #
                # We have the name of the relation (alias), the query is added as a query plan to
                # the parent plan.
                if subquery["alias"] is None:
                    from opteryx.exceptions import UnnamedSubqueryError

                    raise UnnamedSubqueryError(
                        "Ensure you provide a name for all subqueries in FROM or JOIN clauses by using AS)."
                    )

                subquery_step = LogicalPlanNode(node_type=LogicalPlanStepType.Subquery)
                subquery_step.alias = subquery["alias"]["name"]["value"]
                step_id = random_string()
                sub_plan.add_node(step_id, subquery_step)

                subquery_plan = plan_query(subquery["subquery"])
                exit_node = subquery_plan.get_exit_points()[0]
                subquery_step.columns = subquery_plan[exit_node].columns
                subquery_plan.remove_node(exit_node, heal=True)

                sub_plan += subquery_plan
                subquery_entry_id = subquery_plan.get_exit_points()[0]
                sub_plan.add_edge(subquery_entry_id, step_id)

                root_node = step_id
                relation["step_id"] = step_id
            else:
                # VALUES nodes are where the relation is defined within the SQL statement.
                # e.g. SELECT * FROM (VALUES(1),(2)) AS numbers (number)
                #
                # We have the name of the relation (alias), the column names (columns) and the
                # values in each row (values)
                values_step = LogicalPlanNode(
                    node_type=LogicalPlanStepType.FunctionDataset, function="VALUES"
                )
                values_step.alias = subquery["alias"]["name"]["value"]
                values_step.columns = tuple(col["value"] for col in subquery["alias"]["columns"])
                values_step.values = [
                    tuple(logical_planner_builders.build(value) for value in row)
                    for row in subquery["subquery"]["body"]["Values"]["rows"]
                ]
                step_id = random_string()
                sub_plan.add_node(step_id, values_step)
                root_node = step_id
        else:  # pragma: no cover
            raise NotImplementedError(relation["relation"]["Derived"])
    elif is_view(relation["relation"]["Table"]["name"][0]["value"]):
        # We're a view, we need to add it to the plan as a subquery
        view_name = relation["relation"]["Table"]["name"][0]["value"]
        sub_plan = view_as_plan(view_name=view_name)
        plan_head = sub_plan.get_exit_points()[0]

        # Replace the exit node with a subquery node
        # We call the subquery the name of the view if we don't have an alias
        # and we use the colums from the exit node
        subquery_node = LogicalPlanNode(node_type=LogicalPlanStepType.Subquery)
        subquery_node.alias = (
            view_name
            if relation["relation"]["Table"]["alias"] is None
            else relation["relation"]["Table"]["alias"]["name"]["value"]
        )
        subquery_node.columns = sub_plan[plan_head].columns
        sub_plan[plan_head] = subquery_node
        root_node = plan_head

        # DEBUG: log (f"VIEW PLAN")
        # DEBUG: log (sub_plan)

    elif relation["relation"]["Table"]["args"]:
        # If we have args, we're a function dataset (like FAKE or UNNEST)
        function = relation["relation"]["Table"]
        function_name = function["name"][0]["value"].upper()

        if function["alias"] is None:
            from opteryx.exceptions import UnnamedColumnError

            raise UnnamedColumnError(
                f"Column or Relation created by {function_name} has no name, use AS to give it a name."
            )

        function_step = LogicalPlanNode(
            node_type=LogicalPlanStepType.FunctionDataset, function=function_name
        )
        if function_name == "UNNEST":
            function_step.alias = f"$unnest-{random_string(6)}"
            function_step.relation = function_step.alias
            function_step.unnest_target = function["alias"]["name"]["value"]
        else:
            function_step.alias = function["alias"]["name"]["value"]

        function_step.args = [
            logical_planner_builders.build(arg) for arg in function["args"]["args"]
        ]
        function_step.columns = tuple(col["value"] for col in function["alias"]["columns"])

        step_id = random_string()
        sub_plan.add_node(step_id, function_step)
        root_node = step_id
        relation["step_id"] = step_id
    else:
        # SCAN nodes are where we read relations; these can be from memory, disk or a remote
        # system. This has many physical implementations but at this point all we have is the
        # name/location of the relation (relation), what the relation is called inside the
        # query (alias) and if there are any hints (hints)
        from_step = LogicalPlanNode(node_type=LogicalPlanStepType.Scan)
        table = relation["relation"]["Table"]
        from_step.relation = ".".join(part["value"] for part in table["name"])
        from_step.alias = (
            from_step.relation if table["alias"] is None else table["alias"]["name"]["value"]
        )
        from_step.hints = [hint["Identifier"]["value"] for hint in table["with_hints"]]
        from_step.start_date = table.get("start_date")
        from_step.end_date = table.get("end_date")
        step_id = random_string()
        sub_plan.add_node(step_id, from_step)

        root_node = step_id
        relation["step_id"] = step_id

    # joins
    _joins = relation.get("joins", [])
    for join in _joins:
        # this is the convention: select * from LEFT join RIGHT

        join_step = process_join_tree(join)

        right_node_id, right_plan = create_node_relation(join)

        # add the left and right relation names - we sometimes need these later
        join_step.left_relation_names = get_subplan_schemas(sub_plan)
        join_step.right_relation_names = get_subplan_schemas(right_plan)

        # add the right side of the join
        sub_plan += right_plan

        join_step_id = random_string()
        sub_plan.add_node(join_step_id, join_step)

        # add the from table as the left side of the join
        sub_plan.add_edge(root_node, join_step_id, "left")
        sub_plan.add_edge(right_node_id, join_step_id, "right")

        root_node = join_step_id

    return root_node, sub_plan


def analyze_query(statement) -> LogicalPlan:
    root_node = "Analyze"
    plan = LogicalPlan()

    from_step = LogicalPlanNode(node_type=LogicalPlanStepType.Scan)
    table = statement[root_node]["table_name"]
    from_step.relation = ".".join(part["value"] for part in table)
    from_step.alias = from_step.relation
    from_step.start_date = table[0].get("start_date")
    from_step.end_date = table[0].get("end_date")
    step_id = random_string()
    plan.add_node(step_id, from_step)

    metadata_step = LogicalPlanNode(node_type=LogicalPlanStepType.MetadataWriter)
    previous_step_id, step_id = step_id, random_string()
    plan.add_node(step_id, metadata_step)
    plan.add_edge(previous_step_id, step_id)

    return plan
    # write manifest


def plan_execute_query(statement) -> LogicalPlan:
    import orjson

    from opteryx.exceptions import SqlError
    from opteryx.planner.ast_rewriter import do_ast_rewriter
    from opteryx.planner.logical_planner import do_logical_planning_phase
    from opteryx.planner.sql_rewriter import do_sql_rewrite
    from opteryx.third_party import sqloxide
    from opteryx.utils import sql

    def build_parm(node):
        if "BinaryOp" in node:
            op = node["BinaryOp"]
            if "op" in op and op["op"] == "Eq":
                return (
                    logical_planner_builders.build(op["left"]).value,
                    logical_planner_builders.build(op["right"]).value,
                )
        from opteryx.exceptions import ParameterError

        raise ParameterError("EXECUTE paramters must be named, e.g. 'paramter=value'")

    # the parser allows USING, but we want EXECUTE function (parmeters)
    if statement["Execute"].get("using"):
        raise UnsupportedSyntaxError(
            "EXECUTE does not support USING syntax, please provide parameters in parenthesis."
        )

    statement_name = statement["Execute"]["name"]["value"].upper()
    parameters = dict(build_parm(p) for p in statement["Execute"]["parameters"])
    try:
        with open("prepared_statements.json", "r") as ps:
            prepared_statatements = {str(k).upper(): v for k, v in orjson.loads(ps.read()).items()}
    except (OSError, ValueError):
        prepared_statatements = {}

    # we have an inbuilt statements as a fallback
    prepared_statatements["PLANETS_BY_ID"] = {
        "statement": "SELECT * FROM $planets WHERE id = :id",
        "parameters": [{"id": "name", "type": "INTEGER"}],
    }
    prepared_statatements["VERSION"] = {
        "statement": "SELECT version()",
        "parameters": [],
    }
    if statement_name not in prepared_statatements:
        raise SqlError(f"Unable to EXECUTE prepared statement, '{statement_name}' not defined.")
    operation = prepared_statatements[statement_name]["statement"]

    operation = sql.remove_comments(operation)
    operation = sql.clean_statement(operation)
    statements = sql.split_sql_statements(operation)
    if len(statements) > 1:
        raise UnsupportedSyntaxError("EXECUTE cannot run multi-part and batch prepared statements.")

    clean_sql, temporal_filters = do_sql_rewrite(operation)
    try:
        parsed_statements = sqloxide.parse_sql(clean_sql, dialect="mysql")
    except ValueError as parser_error:
        raise SqlError(parser_error) from parser_error

    parsed_statements = do_ast_rewriter(
        parsed_statements,
        temporal_filters=temporal_filters,
        parameters=parameters,
        connection=None,
    )
    return list(do_logical_planning_phase(parsed_statements))[0][0]


def plan_explain(statement) -> LogicalPlan:
    plan = LogicalPlan()
    explain_node = LogicalPlanNode(node_type=LogicalPlanStepType.Explain)
    explain_node.analyze = statement["Explain"]["analyze"]
    explain_node.format = statement["Explain"]["format"]

    explain_id = random_string()
    plan.add_node(explain_id, explain_node)

    sub_plan = plan_query(statement=statement["Explain"]["statement"])
    sub_plan_id = sub_plan.get_exit_points()[0]
    plan += sub_plan
    plan.add_edge(sub_plan_id, explain_id)

    return plan


def plan_query(statement):
    """ """

    root_node = statement
    if "Query" in root_node:
        root_node = root_node["Query"]

    # union?
    if "SetOperation" in root_node["body"]:
        set_operation = root_node["body"]["SetOperation"]

        if set_operation["op"] == "Union":
            set_op_node = LogicalPlanNode(node_type=LogicalPlanStepType.Union)
        else:
            raise UnsupportedSyntaxError(f"Unsupported SET operator '{set_operation['op']}'")

        set_op_node.modifier = (
            None if set_operation["set_quantifier"] == "None" else set_operation["set_quantifier"]
        )
        step_id = random_string()
        plan = LogicalPlan()
        plan.add_node(step_id, set_op_node)
        head_nid = step_id

        left_plan = inner_query_planner(set_operation["left"])
        plan += left_plan
        subquery_entry_id = left_plan.get_exit_points()[0]
        plan.add_edge(subquery_entry_id, step_id)
        # remove the exit node
        plan.remove_node(subquery_entry_id, heal=True)

        right_plan = inner_query_planner(set_operation["right"])
        plan += right_plan
        subquery_entry_id = right_plan.get_exit_points()[0]
        plan.add_edge(subquery_entry_id, step_id)
        # remove the exit node
        plan.remove_node(subquery_entry_id, heal=True)

        # UNION ALL
        if set_op_node.modifier != "All":
            distinct = LogicalPlanNode(node_type=LogicalPlanStepType.Distinct)
            head_nid, step_id = step_id, random_string()
            plan.add_node(step_id, distinct)
            plan.add_edge(head_nid, step_id)

        # limit/offset
        _limit = root_node.get("limit")
        _offset = root_node.get("offset")
        if _offset:
            _offset = _offset.get("value")
        if _limit or _offset:
            limit_step = LogicalPlanNode(node_type=LogicalPlanStepType.Limit)
            limit_step.limit = (
                None if _limit is None else logical_planner_builders.build(_limit).value
            )
            limit_step.offset = (
                None if _offset is None else logical_planner_builders.build(_offset).value
            )
            head_nid, step_id = step_id, random_string()
            plan.add_node(step_id, limit_step)
            if head_nid is not None:
                plan.add_edge(head_nid, step_id)

        # add the exit node
        exit_node = LogicalPlanNode(node_type=LogicalPlanStepType.Exit)
        _projection_nodes = [
            left_plan[nid]
            for nid in left_plan.nodes()
            if left_plan[nid].node_type in (LogicalPlanStepType.Project,)
        ]
        columns = [LogicalPlanNode(NodeType.WILDCARD, value=(None,))]
        if _projection_nodes:
            columns = _projection_nodes[0].columns
        exit_node.columns = columns
        head_nid, step_id = step_id, random_string()
        plan.add_node(step_id, exit_node)
        if head_nid is not None:
            plan.add_edge(head_nid, step_id)

        set_op_node.columns = columns
        set_op_node.left_relation_names = get_subplan_schemas(left_plan)
        set_op_node.right_relation_names = get_subplan_schemas(right_plan)

        return plan

    # we do some minor AST rewriting
    root_node["body"]["limit"] = root_node.get("limit", None)
    root_node["body"]["offset"] = (root_node.get("offset") or {}).get("value")
    root_node["body"]["order_by"] = root_node.get("order_by", None)

    planned_query = inner_query_planner(root_node["body"])

    # DEBUG: log ("LOGICAL PLAN")
    # DEBUG: log (planned_query.draw())

    return planned_query


def plan_set_variable(statement):
    root_node = "SetVariable"
    plan = LogicalPlan()
    set_step = LogicalPlanNode(
        node_type=LogicalPlanStepType.Set,
        variable=extract_variable(statement[root_node]["variables"]["One"]),
        value=extract_value(statement[root_node]["value"]),
    )
    plan.add_node(random_string(), set_step)
    return plan


def plan_show_columns(statement):
    root_node = "ShowColumns"
    plan = LogicalPlan()

    from_step = LogicalPlanNode(node_type=LogicalPlanStepType.Scan)
    table = statement[root_node]["table_name"]
    from_step.relation = ".".join(part["value"] for part in table)
    from_step.alias = from_step.relation
    from_step.start_date = table[0].get("start_date")
    from_step.end_date = table[0].get("end_date")
    step_id = random_string()
    plan.add_node(step_id, from_step)

    show_step = LogicalPlanNode(node_type=LogicalPlanStepType.ShowColumns)
    show_step.extended = statement[root_node]["extended"]
    show_step.full = statement[root_node]["full"]
    show_step.relation = from_step.relation
    previous_step_id, step_id = step_id, random_string()
    plan.add_node(step_id, show_step)
    plan.add_edge(previous_step_id, step_id)

    _filter = statement[root_node]["filter"]
    if _filter:
        filter_node = LogicalPlanNode(node_type=LogicalPlanStepType.Filter)
        filter_node.condition = extract_simple_filter(_filter, "name")
        previous_step_id, step_id = step_id, random_string()
        plan.add_node(step_id, filter_node)
        plan.add_edge(previous_step_id, step_id)
        raise UnsupportedSyntaxError("Unable to filter colmns in SHOW COLUMNS")

    return plan


def plan_show_variable(statement):
    root_node = "ShowVariable"
    plan = LogicalPlan()
    show_step = LogicalPlanNode(node_type=LogicalPlanStepType.Show)
    show_step.items = extract_variable(statement[root_node]["variable"])
    plan.add_node(random_string(), show_step)
    return plan


def plan_show_variables(statement):
    root_node = "ShowVariables"
    plan = LogicalPlan()

    read_step = LogicalPlanNode(node_type=LogicalPlanStepType.Scan)
    read_step.relation = "$variables"
    step_id = random_string()
    plan.add_node(step_id, read_step)

    predicate = statement[root_node]["filter"]
    if predicate is not None:
        operator = next(iter(predicate))
        select_step = LogicalPlanNode(
            node_type=LogicalPlanStepType.Filter,
            condition=Node(
                node_type=NodeType.COMPARISON_OPERATOR,
                value=operator,
                left=Node(node_type=NodeType.IDENTIFIER, value="name"),
                right=Node(
                    node_type=NodeType.LITERAL,
                    type=OrsoTypes.VARCHAR,
                    value=predicate[operator],
                ),
            ),
        )
        previous_step_id, step_id = step_id, random_string()
        plan.add_node(step_id, select_step)
        plan.add_edge(previous_step_id, step_id)
        raise UnsupportedSyntaxError("Cannot filter by Variable Names")

    exit_step = LogicalPlanNode(node_type=LogicalPlanStepType.Exit)
    exit_step.columns = [Node(node_type=NodeType.WILDCARD)]  # We are always SELECT *
    previous_step_id, step_id = step_id, random_string()
    plan.add_node(step_id, exit_step)
    plan.add_edge(previous_step_id, step_id)

    return plan


QUERY_BUILDERS = {
    "Analyze": analyze_query,
    "Execute": plan_execute_query,
    "Explain": plan_explain,
    "Query": plan_query,
    "SetVariable": plan_set_variable,
    "ShowColumns": plan_show_columns,
    #    "ShowCreate": show_create_query,
    #    "ShowFunctions": show_functions_query,
    "ShowVariable": plan_show_variable,  # generic SHOW handler
    "ShowVariables": plan_show_variables,
    # "Use": plan_use
}


def do_logical_planning_phase(parsed_statements) -> Generator:
    # The sqlparser ast is an array of asts
    for parsed_statement in parsed_statements:
        statement_type = next(iter(parsed_statement))
        if statement_type not in QUERY_BUILDERS:
            from opteryx.exceptions import UnsupportedSyntaxError

            raise UnsupportedSyntaxError(
                f"Version 2 Planner does not support '{statement_type}' type queries yet."
            )
        # CTEs are Common Table Expressions, they're variations of subqueries
        ctes = extract_ctes(parsed_statement, inner_query_planner)
        yield QUERY_BUILDERS[statement_type](parsed_statement), parsed_statement, ctes



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains support functions for working with PyArrow
"""

from typing import Iterator
from typing import List
from typing import Optional
from typing import Union

import pyarrow

INTERNAL_BATCH_SIZE = 500


def limit_records(
    morsels: Iterator[pyarrow.Table], limit: Optional[int] = None, offset: int = 0
) -> Optional[Iterator[pyarrow.Table]]:
    """
    Cycle over an iterable of morsels, limiting the response to a given
    number of records with an optional offset.
    """
    remaining_rows = limit if limit is not None else float("inf")
    rows_left_to_skip = max(0, offset)
    at_least_one = False

    for morsel in morsels:
        if rows_left_to_skip > 0:
            if rows_left_to_skip >= morsel.num_rows:
                rows_left_to_skip -= morsel.num_rows
                continue
            else:
                morsel = morsel.slice(
                    offset=rows_left_to_skip, length=morsel.num_rows - rows_left_to_skip
                )
                rows_left_to_skip = 0

        if morsel.num_rows > 0:
            if morsel.num_rows < remaining_rows:
                yield morsel
                at_least_one = True
            else:
                yield morsel.slice(offset=0, length=remaining_rows)
                at_least_one = True

        if not at_least_one:
            # make sure we return at least an empty morsel from this function
            yield morsel.slice(offset=0, length=0)
            at_least_one = True

        remaining_rows -= morsel.num_rows
        if remaining_rows <= 0:
            break

    if not remaining_rows:
        return None


def restore_null_columns(removed, table):
    for column in removed:  # pragma: no cover
        table = table.append_column(column, pyarrow.array([None] * table.num_rows))
    return table


def post_read_projector(table: pyarrow.Table, columns: list) -> pyarrow.Table:
    """
    This is the near-read projection for data sources that the projection can't be
    done as part of the read.
    """
    if not columns:
        # this should happen when there's no relation in the query
        return table

    schema_columns = set(table.column_names)

    # Using a dictionary to map all_names to the projection_column's name for quick lookup
    name_mapping = {
        name: projection_column.schema_column.name
        for projection_column in columns
        for name in projection_column.schema_column.all_names
    }

    columns_to_keep = [
        schema_column for schema_column in schema_columns if schema_column in name_mapping
    ]
    column_names = [name_mapping[schema_column] for schema_column in columns_to_keep]

    table = table.select(columns_to_keep)
    return table.rename_columns(column_names)


def align_tables(
    source_table: pyarrow.Table,
    append_table: pyarrow.Table,
    source_indices: Union[List[int], pyarrow.Array],
    append_indices: Union[List[int], pyarrow.Array],
) -> pyarrow.Table:
    """
    Aligns two tables based on provided indices, ensuring that the resulting table
    contains columns from both source_table and append_table.

    Parameters:
        source_table: The pyarrow.Table to align.
        append_table: The pyarrow.Table to align with the source_table.
        source_indices: The indices to take from the source_table.
        append_indices: The indices to take from the append_table.

    Returns:
        A pyarrow.Table with aligned columns and data.
    """

    # If either source_indices or append_indices is empty, return an empty table with a combined schema
    if len(source_indices) == 0 or len(append_indices) == 0:
        combined_schema = pyarrow.schema(
            [
                *source_table.schema,
                *[
                    field
                    for field in append_table.schema
                    if field.name not in source_table.schema.names
                ],
            ]
        )
        empty_arrays = [pyarrow.array([]) for field in combined_schema]
        return pyarrow.Table.from_arrays(empty_arrays, schema=combined_schema)

    # Convert indices to PyArrow arrays for efficient null checking
    source_indices_array = pyarrow.array(source_indices)
    append_indices_array = pyarrow.array(append_indices)

    # Check if all source_indices are nulls
    if source_indices_array.null_count == len(source_indices):
        null_columns = [
            pyarrow.nulls(len(source_indices), type=field.type) for field in source_table.schema
        ]
        aligned_table = pyarrow.Table.from_arrays(null_columns, schema=source_table.schema)
    else:
        # Take rows from source_table based on source_indices
        aligned_table = source_table.take(source_indices_array)

    # Set of column names from source_table for quick lookup
    source_column_names = set(source_table.column_names)

    # Check if all append_indices are nulls
    append_is_all_nulls = append_indices_array.null_count == len(append_indices)

    new_columns = []
    new_fields = []

    # Collect all columns that need to be appended in one go
    for column_name, column_field in zip(append_table.column_names, append_table.schema):
        if column_name not in source_column_names:
            if append_is_all_nulls:
                # Create a column of nulls if all append indices are null
                null_column = pyarrow.nulls(len(aligned_table), type=column_field.type)
                new_columns.append(null_column)
            else:
                # Take the corresponding rows from append_table
                column_data = append_table.column(column_name).take(append_indices_array)
                new_columns.append(column_data)
            new_fields.append(column_field)

    # Combine original and new columns
    aligned_columns = aligned_table.columns + new_columns
    combined_schema = pyarrow.schema([*source_table.schema, *new_fields])

    # Create the final aligned table with all columns at once
    aligned_table = pyarrow.Table.from_arrays(aligned_columns, schema=combined_schema)

    return aligned_table



"""
travers

(C) 2023 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import orjson

from opteryx.exceptions import MissingDependencyError


def print_tree_inner(tree, prefix="", last=True):
    """
    Prints a nested dictionary as an ascii tree
    """

    yield prefix
    if last:
        yield "└─ "
        prefix += "   "
    else:
        yield "├─ "
        prefix += "│  "

    yield str(tree["node"]) + "\n"

    # Recursively print the children
    count = len(tree["children"])
    for i, child in enumerate(tree["children"]):
        last = i == count - 1
        yield from print_tree_inner(child, prefix, last)


class Graph(object):
    """
    Graph object, optimized for traversal.

    Edges are stored in a dictionary, the key is the source node to speed up
    finding outgoing edges. The Edges only have three pieces of data:
        - the source node (the key)
        - the target node
        - the relationship
    The target and the relationship are stored as a tuple, the edge dictionary
    stores lists of tuples.

    """

    __slots__ = ("_nodes", "_edges")

    def __init__(self):
        """
        Directed Graph.
        """
        self._nodes = {}
        self._edges = {}

    def __bool__(self) -> bool:
        return len(self._nodes) != 0 or len(self._edges) != 0

    def save(self, graph_path):  # pragma: nocover
        """
        Persist a graph to storage. It saves nodes and edges to separate files.

        Parameters:
            graph_path: string
                The folder ?to save the node and edge files to
        """
        path = Path(graph_path)
        path.mkdir(exist_ok=True)

        with open(path / "edges.jsonl", "wb") as edge_file:
            for source, target, relationship in self.edges():
                edge_record = {
                    "source": source,
                    "target": target,
                    "relationship": relationship,
                }
                edge_file.write(orjson.dumps(edge_record) + b"\n")
        with open(path / "nodes.jsonl", "wb") as node_file:
            for nid, attr in self.nodes(data=True):
                node_file.write(orjson.dumps({"nid": nid, "attributes": attr}) + b"\n")

    def add_edge(self, source: str, target: str, relationship: Optional[str] = None):
        """
        Add edge to the graph

        Note:
            This does not create an edge if either node does not already exist.
            This does not create an edge if either node is None.

        Parameters:
            source: string
                The source node
            target: string
                The target node
            relationship: string
                The relationship between the source and target nodes
        """
        if source is None or target is None:
            print("Trying to create edge with undefined nodes")
            return False

        # Check for existing edges and add the new one
        existing_edges = list(self._edges.get(source, ()))

        # Avoid adding duplicate edges
        edge_to_add = (target, relationship)
        if edge_to_add not in existing_edges:
            existing_edges.append(edge_to_add)

        self._edges[source] = tuple(existing_edges)

    def add_node(self, nid: str, node):
        """
        Add node to the graph

        Parameters:
            nid: string
                The Node ID for the node (unique)
            attributes: dictionary (optional)
                The attributes of the node
        """
        self._nodes[nid] = node

    def nodes(self, data=False):
        """
        The nodes which comprise the graph

        Parameters:
            data: boolean (optional)
                if True return the details of the nodes, if False just return
                the list of node IDs

        Returns:
            List
        """
        if data:
            return self._nodes.items()
        return list(self._nodes.keys())

    def edges(self):
        """
        The edges which comprise the graph

        Returns:
            Generator of Tuples of (Source, Target and Relationship)
        """
        for source, records in self._edges.items():
            yield from ((source, target, relationship) for target, relationship in records)

    def breadth_first_search(
        self, source: str, depth: int = 100, reverse: bool = False
    ):  # pragma: nocover
        """
        Search a tree for nodes we can walk to from a given node.

        Parameters:
            source: string
                The node to walk from
            depth: integer
                The maximum distance to walk from source
        Returns:
        """
        from collections import deque

        visited = {source}
        queue = deque([(source, 0)])
        seeker = self.ingoing_edges if reverse else self.outgoing_edges

        traversed_edges = []

        while queue:
            current_node, current_depth = queue.popleft()

            if current_depth < depth:
                for edge in seeker(current_node):
                    _, target, _ = edge

                    # Add the edge to the traversed edges list
                    traversed_edges.append(edge)

                    if target not in visited:
                        visited.add(target)
                        queue.append((target, current_depth + 1))

        return traversed_edges

    def depth_first_search(
        self, node: Optional[str] = None, visited: Optional[set] = None, depth: int = 0
    ):
        """
        Returns a nested dictionary representing the graph as a tree
        """
        if node is None:
            node = self.get_exit_points()[0]

        if visited is None:
            visited = set()

        visited.add(node)

        tree: dict = {
            "type": str(self[node].node_type),
            "node": str(self[node]),
            "name": node,
            "depth": depth,
            "children": [],
        }

        for neighbor, _, relationship in self.ingoing_edges(node):
            if neighbor not in visited:
                child = self.depth_first_search(neighbor, visited, depth + 1)
                child["relationship"] = relationship
                tree["children"].append(child)  # type:ignore

        return tree

    def outgoing_edges(self, source) -> List[Tuple]:
        """
        Get the list of edges traversable from a given node.

        Parameters:
            source: string
                The node to get the outgoing edges for

        Returns:
            Set of Tuples (Source, Target, Relationship)
        """
        return [(source, t, r) for t, r in self._edges.get(source, tuple())]

    def ingoing_edges(self, target) -> List[Tuple]:
        """
        Get the list of edges which can traverse to a given node.

        Parameters:
            target: string
                The node to get the incoming edges for

        Returns:
            Set of Tuples (Source, Target, Relationship)
        """
        return [(s, t, r) for s, t, r in self.edges() if t == target]

    def is_acyclic(self):
        """
        Test if the Graph is acyclic
        """
        # cycle over the graph removing a layer of exits each cycle
        # if we have nodes but no exists, we're cyclic

        # rebuild the edge information
        my_edges = list(self.edges())

        while len(my_edges) > 0:
            # find all of the exits
            sources = {source for source, target, direction in my_edges}
            exits = {target for source, target, direction in my_edges if target not in sources}

            if len(exits) == 0:
                return False

            # remove the exits
            new_edges = [
                (source, target, direction)
                for source, target, direction in my_edges
                if target not in exits
            ]
            my_edges = new_edges
        return True

    def shortest_path(self, start: str, end: str) -> List[str]:
        """
        Compute the shortest path from start to end node.

        Parameters:
            start: string
                The starting node ID
            end: string
                The target node ID

        Returns:
            List of node IDs from start to end node that represent the shortest path.
            Returns an empty list if no path is found.
        """

        from collections import deque

        visited = set()
        queue = deque([(start, [start])])  # Each item in the queue is a tuple (node, path_so_far)

        while queue:
            node, path = queue.popleft()

            if node == end:
                return path  # Found a path to the end node

            if node not in visited:
                visited.add(node)

                for _, neighbor, _ in self.outgoing_edges(node):
                    if neighbor == end:
                        path.append(neighbor)
                        return path
                    if neighbor not in visited:
                        new_path = list(path)
                        new_path.append(neighbor)
                        queue.append((neighbor, new_path))

        return []  # No path found

    def get_entry_points(self):
        """
        Get nodes in the Graph with no incoming edges.
        """
        if len(self._nodes) == 1:
            return list(self._nodes.keys())

        edges = list(self.edges())
        targets = {target for _, target, _ in edges}
        sources = {source for source, _, _ in edges}
        return sorted(sources - targets)

    def get_exit_points(self):
        """
        Get nodes in the Graph with no outgoing edges.
        """
        if len(self._nodes) == 1:
            return list(self._nodes.keys())

        edges = list(self.edges())
        targets = {target for _, target, _ in edges}
        sources = set(self._edges.keys())
        return sorted(targets - sources)

    def remove_node(self, nid, heal: bool = False):
        """
        Remove a node.

        Parameters:
            heal: boolean
                Join the incoming and outgoing connections for the removed node
                to each other to keep the Graph intact
        """

        # remove the node
        self._nodes.pop(nid, None)

        if heal:
            # link the nodes each side of the node being removed
            out_going = self.outgoing_edges(nid)
            in_coming = self.ingoing_edges(nid)

            # remove edges where the node is the source
            if nid in self._edges:
                del self._edges[nid]

            # remove the edges where the node is the target
            for source, records in self._edges.items():
                self._edges[source] = [
                    (target, relationship) for target, relationship in records if target != nid
                ]
            self._edges = {k: v for k, v in self._edges.items() if len(v) > 0}

            # wire up the old incoming and outgoing nodes, cartesian style
            for out_nid in out_going:
                for in_nid in in_coming:
                    self.add_edge(in_nid[0], out_nid[1], in_nid[1])  # type:ignore

    def remove_edge(self, source, target, relationship):
        """
        Remove an edge from the graph.
        Args:
        - source (str): The source node of the edge.
        - target (str): The target node of the edge.
        - relationship (str): The relationship label of the edge.
        """
        if source not in self._edges:
            return
        edge_to_remove = (target, relationship)
        if source in self._edges and edge_to_remove in self._edges[source]:
            working_set = list(self._edges[source])
            working_set.remove(edge_to_remove)
            self._edges[source] = tuple(working_set)
            if not self._edges[source]:  # If no edges left for the source
                del self._edges[source]

    def insert_node_before(self, nid, node, before_nid):
        """rewrite the plan putting the new node before a given node"""
        # add the new node to the plan
        self.add_node(nid, node)
        # change all the edges that were going into the old nid to the new one

        for source, records in self._edges.items():
            new_records = []
            for target, relationship in records:
                if target != before_nid:
                    new_records.append(
                        (
                            target,
                            relationship,
                        )
                    )
                else:
                    new_records.append(
                        (
                            nid,
                            relationship,
                        )
                    )
                self._edges[source] = new_records
        # add an edge from the new nid to the old one
        self.add_edge(nid, before_nid)

    def insert_node_after(self, nid, node, after_nid):
        """rewrite the plan putting the new node after a given node"""
        # add the new node to the plan
        self.add_node(nid, node)
        # change all the edges that were coming from the old nid to the new one
        if after_nid in self._edges:
            self._edges[nid] = self._edges.pop(after_nid)
        # add an edge from the new nid to the old one
        self.add_edge(after_nid, nid)

    def to_networkx(self):  # pragma: nocover
        """
        Convert a travers graph to a NetworkX graph
        """
        try:
            import networkx as nx  # type:ignore
        except ImportError as err:  # pragma: no cover
            raise MissingDependencyError(err.name) from err

        g = nx.DiGraph()
        for s, t, r in self.edges():
            g.add_edge(s, t, relationship=r)
        for node, attribs in self.nodes(True):
            g.add_node(node, **attribs)
        return g

    def epitomize(self):  # pragma: nocover
        """
        Summarize a Graph by reducing to only the node_types and relationships
        """
        g = Graph()
        for s, t, r in self.edges():
            node1 = self[s]
            node2 = self[t]
            if node1 and node2:
                g.add_edge(node1.get("node_type"), node2.get("node_type"), r)
            if node1:
                g.add_node(node1.get("node_type"), {"node_type": node1.get("node_type")})
            if node2:
                g.add_node(node2.get("node_type"), {"node_type": node2.get("node_type")})
        return g

    def __repr__(self):
        return f"Graph - {len(list(self.nodes()))} nodes, {len(list(self.edges()))} edges"

    def __len__(self):
        return len(list(self.nodes()))

    def __getitem__(self, nid):
        return self._nodes.get(nid, None)

    def __setitem__(self, nid, node):
        if nid not in self._nodes:
            raise ValueError("Cannot create nodes with [] syntax")
        self._nodes[nid] = node

    def __add__(self, other):
        self._edges.update(other._edges)
        self._nodes.update(other._nodes)
        return self

    def __contains__(self, nid: str) -> bool:
        return nid in self._nodes

    def draw(self):
        tree = self.depth_first_search()
        return "".join(print_tree_inner(tree))

    def copy(self) -> "Graph":
        """
        Intelligently make a copy of this Graph, handling situations where Deepcopy
        does not work.
        """
        import copy

        def _inner_copy(obj: Any) -> Any:
            """
            Create an independent inner copy of the given object.

            Parameters:
                obj: Any
                    The object to be deep copied.

            Returns:
                Any: The new, independent deep copy.
            """
            if isinstance(obj, list):
                return [_inner_copy(item) for item in obj]
            if isinstance(obj, tuple):
                return tuple(_inner_copy(item) for item in obj)
            if isinstance(obj, set):
                return {_inner_copy(item) for item in obj}
            if isinstance(obj, dict):
                return {key: _inner_copy(value) for key, value in obj.items()}
            if hasattr(obj, "copy"):
                return obj.copy()
            try:
                return copy.deepcopy(obj)
            except:
                return obj

        graph = Graph()
        graph._nodes = _inner_copy(self._nodes)
        graph._edges = self.copy_edges()

        return graph

    def copy_edges(self):
        """
        Creates an independent copy of the edges in the graph.

        Returns:
            A new dictionary representing the edges in the graph.
        """
        new_edges: dict = {}
        for source, target, relationship in self.edges():
            if source not in new_edges:
                new_edges[source] = []
            new_edges[source].append((target, relationship))
        return new_edges

    def trace_to_root(self, nid: str) -> list:
        """
        Traces the path from this node to the root of the tree, recording each node along the way.

        Args:
            tree: The tree structure containing the nodes and edges.
        """
        route = []
        current_node_id = nid
        while True:
            # Get the node before the current node
            outgoing_edges = self.outgoing_edges(current_node_id)
            if not outgoing_edges:
                break  # Reached the root

            # Assuming the first element of the first tuple in ingoing_edges is the previous node's ID
            previous_node_id = outgoing_edges[0][1]

            # Record this node in the chain
            route.append(previous_node_id)

            # Move to the previous node
            current_node_id = previous_node_id

        return route



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
~~~
                      ┌───────────┐
                      │   USER    │
         ┌────────────┤           ◄────────────┐
         │SQL         └───────────┘            │
  ───────┼─────────────────────────────────────┼──────
         │                                     │
   ┌─────▼─────┐                               │
   │ SQL       │                               │
   │ Rewriter  │                               │
   └─────┬─────┘                               │
         │SQL                                  │Results
   ┌─────▼─────┐                         ┌─────┴─────┐
   │           │                         │           │
   │ Parser    │                         │ Executor  │
   └─────┬─────┘                         └─────▲─────┘
         │AST                                  │Plan
   ┌─────▼─────┐      ┌───────────┐      ┌─────┴─────┐
   │ AST       │      │           │      │ Physical  │
   │ Rewriter  │      │ Catalogue │      │ Planner   │
   └─────┬─────┘      └───────────┘      └─────▲─────┘
         │AST               │Schemas           │Plan
   ┌─────▼─────┐      ┌─────▼─────┐      ┌─────┴─────┐
   │ Logical   │ Plan │           │ Plan │           │
   │   Planner ├──────► Binder    ├──────► Optimizer │
   └───────────┘      └───────────┘      └───────────┘

~~~
"""

import time
from typing import Any
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import Union

from opteryx import config
from opteryx.models import ExecutionTree

PROFILE_LOCATION = config.PROFILE_LOCATION


def query_planner(
    operation: str,
    parameters: Union[Iterable, Dict, None],
    connection,
    qid: str,
    statistics,
) -> Generator[ExecutionTree, Any, Any]:
    from opteryx.exceptions import SqlError
    from opteryx.models import QueryProperties
    from opteryx.planner.ast_rewriter import do_ast_rewriter
    from opteryx.planner.binder import do_bind_phase
    from opteryx.planner.cost_based_optimizer import do_cost_based_optimizer
    from opteryx.planner.logical_planner import LogicalPlan
    from opteryx.planner.logical_planner import do_logical_planning_phase
    from opteryx.planner.sql_rewriter import do_sql_rewrite
    from opteryx.planner.temporary_physical_planner import create_physical_plan
    from opteryx.third_party import sqloxide

    # SQL Rewriter extracts temporal filters
    start = time.monotonic_ns()
    clean_sql, temporal_filters = do_sql_rewrite(operation)
    statistics.time_planning_sql_rewriter += time.monotonic_ns() - start

    params: Union[list, dict, None] = None
    if parameters is None:
        params = []
    elif isinstance(parameters, dict):
        params = parameters.copy()
    else:
        params = [p for p in parameters or []]

    # Parser converts the SQL command into an AST
    try:
        parsed_statements = sqloxide.parse_sql(clean_sql, dialect="mysql")
    except ValueError as parser_error:
        raise SqlError(parser_error) from parser_error
    # AST Rewriter adds temporal filters and parameters to the AST
    start = time.monotonic_ns()
    parsed_statements = do_ast_rewriter(
        parsed_statements,
        temporal_filters=temporal_filters,
        parameters=params,
        connection=connection,
    )
    statistics.time_planning_ast_rewriter += time.monotonic_ns() - start

    logical_plan: LogicalPlan = None
    ast: dict = {}

    # Logical Planner converts ASTs to logical plans
    for logical_plan, ast, ctes in do_logical_planning_phase(parsed_statements):  # type: ignore
        # check user has permission for this query type
        query_type = next(iter(ast))
        if query_type not in connection.permissions:
            from opteryx.exceptions import PermissionsError

            raise PermissionsError(
                f"User does not have permission to execute '{query_type}' queries."
            )

        # The Binder adds schema information to the logical plan
        start = time.monotonic_ns()
        bound_plan = do_bind_phase(
            logical_plan,
            connection=connection.context,
            qid=qid,
            # common_table_expressions=ctes,
        )
        statistics.time_planning_binder += time.monotonic_ns() - start

        start = time.monotonic_ns()
        optimized_plan = do_cost_based_optimizer(bound_plan)
        statistics.time_planning_optimizer += time.monotonic_ns() - start

        # before we write the new optimizer and execution engine, convert to a V1 plan
        start = time.monotonic_ns()
        query_properties = QueryProperties(qid=qid, variables=connection.context.variables)
        physical_plan = create_physical_plan(optimized_plan, query_properties)
        statistics.time_planning_physical_planner += time.monotonic_ns() - start
        yield physical_plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handle a memoryview like a stream without converting to bytes.

This has the minimal implementation to pass the Opteryx CI, unit and regression
tests - it's not intended to be correct for usage outside Opteryx.
"""

import io
from typing import BinaryIO
from typing import Iterator


class MemoryViewStream(BinaryIO):
    def __init__(self, mv: memoryview):
        self.mv = mv
        self.offset = 0
        self._closed = False

    def read(self, n=-1) -> bytes:
        if n < 0 or n + self.offset > len(self.mv):
            n = len(self.mv) - self.offset
        result = self.mv[self.offset : self.offset + n]
        self.offset += n
        return result.tobytes()

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:  # Absolute file positioning
            self.offset = min(max(offset, 0), len(self.mv))
        elif whence == 1:  # Seek relative to the current position
            self.offset = min(max(self.offset + offset, 0), len(self.mv))
        elif whence == 2:  # Seek relative to the file's end
            self.offset = min(max(len(self.mv) + offset, 0), len(self.mv))
        return self.offset

    def tell(self) -> int:
        return self.offset

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True

    def close(self):
        self._closed = True

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def mode(self) -> str:  # pragma: no cover
        return "rb"

    def __enter__(self) -> BinaryIO:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self) -> Iterator:
        return iter(self.mv)

    def __next__(self) -> bytes:  # pragma: no cover
        if self.offset >= len(self.mv):
            raise StopIteration()
        self.offset += 1
        return bytes([self.mv[self.offset]])

    def fileno(self) -> int:  # pragma: no cover
        return -1

    def flush(self) -> None:  # pragma: no cover
        raise io.UnsupportedOperation()

    def isatty(self) -> bool:  # pragma: no cover
        return False

    def readline(self, limit: int = -1):  # pragma: no cover
        raise io.UnsupportedOperation()

    def readlines(self, hint: int = -1) -> list:  # pragma: no cover
        raise io.UnsupportedOperation()

    def truncate(self):  # pragma: no cover
        raise io.UnsupportedOperation()

    def write(self):  # pragma: no cover
        raise io.UnsupportedOperation()

    def writelines(self):  # pragma: no cover
        raise io.UnsupportedOperation()



# distutils: language = c++
# cython: language_level=3

from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.vector cimport vector
from libc.stdint cimport int64_t, uint8_t
from libcpp.pair cimport pair

cimport cython
cimport numpy as cnp

import numpy
import pyarrow

cdef class HashTable:
    cdef public unordered_map[int64_t, vector[int64_t]] hash_table

    def __init__(self):
        self.hash_table = unordered_map[int64_t, vector[int64_t]]()

    def insert(self, int64_t key, int64_t row_id) -> bool:
        # If the key is already in the hash table, append the row_id to the existing list.
        # Otherwise, create a new list with the row_id.
        if self.hash_table.find(key) != self.hash_table.end():
            self.hash_table[key].push_back(row_id)
            return False
        self.hash_table[key] = vector[int64_t](1, row_id)
        return True

    def get(self, int64_t key) -> list:
        # Return the list of row IDs for the given key, or an empty list if the key is not found.
        if self.hash_table.find(key) != self.hash_table.end():
            return self.hash_table[key]
        return vector[int64_t]()


cdef class HashSet:
    cdef unordered_set[int64_t] c_set

    def __cinit__(self):
        self.c_set = unordered_set[int64_t]()
        self.c_set.reserve(1_048_576)  # try to prevent needing to resize

    cdef inline bint insert(self, int64_t value):
        if self.c_set.find(value) != self.c_set.end():
            return False
        self.c_set.insert(value)
        return True

    cdef inline bint contains(self, int64_t value):
        return self.c_set.find(value) != self.c_set.end()

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline object recast_column(column):
    cdef column_type = column.type

    if pyarrow.types.is_struct(column_type) or pyarrow.types.is_list(column_type):
        return numpy.array([str(a) for a in column.to_pylist()], dtype=numpy.str_)
    return column

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple distinct(table, HashSet seen_hashes=None, list columns=None):
    """
    Perform a distinct operation on the given table using an external SeenHashSet.
    """

    cdef int64_t hashed_value
    cdef int64_t null_hash = hash(None)
    cdef int64_t i = 0
    cdef tuple value_tuple
    cdef object value
    cdef int64_t num_columns

    if seen_hashes is None:
        seen_hashes = HashSet()

    if columns is None:
        columns_of_interest = table.column_names
    else:
        columns_of_interest = columns
    num_columns = len(columns_of_interest)

    cdef list keep = []
    cdef cnp.ndarray values = numpy.array([recast_column(column) for column in table.select(columns_of_interest).itercolumns()], dtype=object)

    if num_columns > 1:
        for i in range(values.shape[1]):
            value_tuple = tuple([v if v == v else None for v in [values[j, i] for j in range(num_columns)]])
            hashed_value = hash(value_tuple)
            if seen_hashes.insert(hashed_value):
                keep.append(i)
    else:
        for i, value in enumerate(values[0]):
            if value != value:
                hashed_value = null_hash
            else:
                hashed_value = hash(value)
            if seen_hashes.insert(hashed_value):
                keep.append(i)

    return (keep, seen_hashes)

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple list_distinct(list values, cnp.ndarray indices, HashSet seen_hashes=None):
    new_indices = []
    new_values = []
    for i, v in enumerate(values):
        if seen_hashes.insert(hash(v)):
            new_values.append(v)
            new_indices.append(indices[i])
    return new_values, new_indices, seen_hashes

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef HashTable hash_join_map(relation, list join_columns):
    """
    Build a hash table for the join operations.

    Parameters:
        relation: The pyarrow.Table to preprocess.
        join_columns: A list of column names to join on.

    Returns:
        A HashTable where keys are hashes of the join column entries and
        values are lists of row indices corresponding to each hash key.
    """
    cdef HashTable ht = HashTable()
    cdef cnp.ndarray[uint8_t, ndim=1] bitmap_array

    # Selecting columns
    cdef int64_t num_rows = relation.num_rows
    cdef int64_t num_columns = len(join_columns)

    # Allocate memory for the combined nulls array
    cdef cnp.ndarray[uint8_t, ndim=1] combined_nulls = numpy.full(num_rows, 1, dtype=numpy.uint8)

    # Process each column to update the combined null bitmap
    cdef int64_t i, col_index
    cdef str column_name
    cdef object column, bitmap_buffer
    cdef uint8_t bit, byte

    for column_name in join_columns:
        column = relation.column(column_name)
        if column.null_count > 0:
            # Get the null bitmap for the current column, ensure it's in a single chunk first
            bitmap_buffer = column.combine_chunks().buffers()[0]
            if bitmap_buffer is not None:
                # Convert the bitmap to uint8
                bitmap_array = numpy.frombuffer(bitmap_buffer, dtype=numpy.uint8)
                for i in range(num_rows):
                    byte = bitmap_array[i // 8]
                    bit = 1 if byte & (1 << (i % 8)) else 0
                    combined_nulls[i] &= bit

    # Determine row indices that have nulls in any of the considered columns
    cdef cnp.ndarray non_null_indices = numpy.nonzero(combined_nulls)[0]

    # Convert selected columns to a numpy array of object dtype, skipping null cols
    cdef cnp.ndarray values_array = numpy.array([relation.column(column).take(non_null_indices) for column in join_columns], dtype=object)
    cdef int64_t hash_value
    cdef tuple value_tuple

    if num_columns > 1:
        for i in range(values_array.shape[1]):
            # Create a tuple of values across the columns for the current row
            value_tuple = tuple(values_array[:, i])
            hash_value = <int64_t>hash(value_tuple)
            ht.insert(hash_value, non_null_indices[i])
    else:
        for i, value in enumerate(values_array[0]):
            hash_value = <int64_t>hash(value)
            ht.insert(hash_value, non_null_indices[i])

    return ht


# cython: language_level=3

import numpy as np  # Required for array allocation

cdef int min3(int x, int y, int z):
    """Utility function to find the minimum of three integers."""
    cdef int m = x
    if y < m:
        m = y
    if z < m:
        m = z
    return m

def levenshtein(str string1, str string2):
    """
    Calculate the Levenshtein distance between two strings.

    Parameters:
        string1 (str): The first string to compare.
        string2 (str): The second string to compare.

    Returns:
        int: The Levenshtein distance between string1 and string2.
    """
    cdef int len1 = len(string1)
    cdef int len2 = len(string2)
    cdef int i, j

    # Allocate a numpy array and create a memory view from it
    cdef int[:] dp = np.zeros((len1 + 1) * (len2 + 1), dtype=np.int32)

    for i in range(len1 + 1):
        for j in range(len2 + 1):
            if i == 0:
                dp[i * (len2 + 1) + j] = j  # First string is empty
            elif j == 0:
                dp[i * (len2 + 1) + j] = i  # Second string is empty
            elif string1[i - 1] == string2[j - 1]:
                dp[i * (len2 + 1) + j] = dp[(i - 1) * (len2 + 1) + (j - 1)]
            else:
                dp[i * (len2 + 1) + j] = 1 + min3(
                    dp[(i - 1) * (len2 + 1) + j],      # Remove
                    dp[i * (len2 + 1) + (j - 1)],      # Insert
                    dp[(i - 1) * (len2 + 1) + (j - 1)] # Replace
                )

    return dp[len1 * (len2 + 1) + len2]



from dataclasses import dataclass
from typing import Optional


@dataclass
class Limit:
    value: Optional[int]  # None indicates unbounded
    inclusive: bool


@dataclass
class ValueRange:
    lower: Limit = None  # Lower limit of the range
    upper: Limit = None  # Upper limit of the range
    untrackable: bool = False

    def update_with_predicate(self, operator: str, value: int):
        if self.untrackable or operator not in ("=", ">=", "<=", ">", "<"):
            self.untrackable = True
            return

        new_limit = Limit(value, inclusive=operator in ("=", ">=", "<="))

        if operator in ("=", ">=", ">"):
            if self.lower is None or (
                new_limit.value > self.lower.value
                or (new_limit.value == self.lower.value and new_limit.inclusive)
            ):
                self.lower = new_limit
        if operator in ("=", "<=", "<"):
            if self.upper is None or (
                new_limit.value < self.upper.value
                or (new_limit.value == self.upper.value and new_limit.inclusive)
            ):
                self.upper = new_limit

    def __bool__(self):
        if self.upper is None or self.lower is None:
            return True
        return self.upper.value >= self.lower.value

    def __str__(self) -> str:
        _range = ""
        if self.untrackable:
            return "Unsupported Conditions"
        if not (self):
            return "Invalid Range"
        if self.lower and self.upper and self.lower.value == self.upper.value:
            return f" = {self.lower.value}"
        if self.lower is not None:
            _range += f" >{'=' if self.lower.inclusive else ''} {self.lower.value}"
        if self.upper is not None:
            _range += f" <{'=' if self.upper.inclusive else ''} {self.upper.value}"
        return _range


# Example usage
initial_range = ValueRange()
initial_range.update_with_predicate(">", 3)
initial_range.update_with_predicate("<", 10)
initial_range.update_with_predicate("=", 7)
print("Range after applying predicates:", initial_range)


def test_initialization():
    vr = ValueRange()
    assert vr.lower is None
    assert vr.upper is None
    assert not vr.untrackable


def test_updates_with_various_predicates():
    vr = ValueRange()
    vr.update_with_predicate(">", 3)
    assert vr.lower == Limit(3, False)
    vr.update_with_predicate("<", 10)
    assert vr.upper == Limit(10, False)


def test_equality_predicate():
    vr = ValueRange()
    vr.update_with_predicate("=", 5)
    assert vr.lower == vr.upper == Limit(5, True)


def test_untrackable_condition():
    vr = ValueRange()
    vr.update_with_predicate("=", 5)
    vr.update_with_predicate("LIKE", "%name")  # This should make the range untrackable
    assert vr.untrackable


def test_predicate_compaction():
    vr = ValueRange()
    # Apply a series of less-than predicates
    vr.update_with_predicate("<", 10)
    vr.update_with_predicate("<", 7)  # Most restrictive
    vr.update_with_predicate("<", 8)
    vr.update_with_predicate("<", 9)

    # After compaction, the upper limit should be '< 7'
    expected_upper_limit = Limit(7, False)  # Assuming exclusive bounds for '<'
    assert (
        vr.upper == expected_upper_limit
    ), f"Expected upper limit to be {expected_upper_limit}, got {vr.upper}"


test_initialization()
test_updates_with_various_predicates()
test_equality_predicate()
test_untrackable_condition()
test_predicate_compaction()
print("okay")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
All of the query types supported by sqlparser-rs
"""

PERMISSIONS = {
    "Analyze",
    "AlterIndex",  # not supported
    "AlterTable",  # not supported
    "Assert",  # not supported
    "Cache",  # not supported
    "Close",  # not supported
    "Commit",  # not supported
    "Comment",  # not supported
    "Copy",  # not supported
    "CreateDatabase",  # not supported
    "CreateFunction",  # not supported
    "CreateIndex",  # not supported
    "CreateRole",  # not supported
    "CreateSchema",  # not supported
    "CreateSequence",  # not supported
    "CreateStage",  # not supported
    "CreateTable",  # not supported
    "CreateView",  # not supported
    "CreateVirtualTable",  # not supported
    "Deallocate",  # not supported
    "Declare",  # not supported
    "Delete",  # not supported
    "Discard",  # not supported
    "Directory",  # not supported
    "Drop",  # not supported
    "DropFunction",  # not supported
    "Execute",
    "Explain",
    "ExplainTable",  # not supported
    "Fetch",  # not supported
    "Grant",  # not supported
    "Insert",  # not supported
    "Kill",  # not supported
    "Merge",  # not supported
    "Msck",  # not supported
    "Prepare",  # not supported
    "Query",
    "Revoke",  # not supported
    "Rollback",  # not supported
    "Savepoint",  # not supported
    "SetNames",  # not supported
    "SetNamesDefault",  # not supported
    "SetRole",  # not supported
    "SetTimeZone",  # not supported
    "SetTransaction",  # not supported
    "SetVariable",
    "ShowCollation",  # not supported
    "ShowColumns",
    "ShowCreate",
    "ShowFunctions",
    "ShowTables",  # not supported
    "ShowVariable",
    "ShowVariables",
    "StartTransaction",  # not supported
    "Truncate",  # not supported
    "UNCache",  # not supported
    "Update",  # not supported
    "Use",
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
MinIo Reader - also works with AWS
"""

import asyncio
import os
from typing import List

import pyarrow
from orso.schema import RelationSchema
from orso.tools import single_item_cache

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import Asynchronous
from opteryx.connectors.capabilities import Cacheable
from opteryx.connectors.capabilities import Partitionable
from opteryx.exceptions import DataError
from opteryx.exceptions import DatasetNotFoundError
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnmetRequirementError
from opteryx.exceptions import UnsupportedFileTypeError
from opteryx.utils import paths
from opteryx.utils.file_decoders import VALID_EXTENSIONS
from opteryx.utils.file_decoders import get_decoder

OS_SEP = os.sep


class AwsS3Connector(BaseConnector, Cacheable, Partitionable, Asynchronous):
    __mode__ = "Blob"
    __type__ = "S3"

    def __init__(self, credentials=None, **kwargs):
        try:
            from minio import Minio  # type:ignore
        except ImportError as err:  # pragma: no cover
            raise MissingDependencyError(err.name) from err

        BaseConnector.__init__(self, **kwargs)
        Partitionable.__init__(self, **kwargs)
        Cacheable.__init__(self, **kwargs)
        Asynchronous.__init__(self, **kwargs)

        # fmt:off
        end_point = kwargs.get("S3_END_POINT", os.environ.get("MINIO_END_POINT"))
        access_key = kwargs.get("S3_ACCESS_KEY", os.environ.get("MINIO_ACCESS_KEY"))
        secret_key = kwargs.get("S3_SECRET_KEY", os.environ.get("MINIO_SECRET_KEY"))
        secure = kwargs.get("S3_SECURE", str(os.environ.get("MINIO_SECURE", "TRUE")).lower() == "true")
        # fmt:on

        if end_point is None:  # pragma: no cover
            raise UnmetRequirementError(
                "MinIo (S3) adapter requires MINIO_END_POINT, MINIO_ACCESS_KEY and MINIO_SECRET_KEY set in environment variables."
            )

        self.minio = Minio(end_point, access_key, secret_key, secure=secure)
        self.dataset = self.dataset.replace(".", OS_SEP)

        # we're going to cache the first blob as the schema and dataset reader
        # sometimes both start here
        self.cached_first_blob = None

    @single_item_cache
    def get_list_of_blob_names(self, *, prefix: str) -> List[str]:
        bucket, object_path, _, _ = paths.get_parts(prefix)
        blobs = self.minio.list_objects(bucket_name=bucket, prefix=object_path, recursive=True)
        blobs = (
            bucket + "/" + blob.object_name for blob in blobs if not blob.object_name.endswith("/")
        )

        return sorted(
            blob for blob in blobs if ("." + blob.split(".")[-1].lower()) in VALID_EXTENSIONS
        )

    def read_dataset(
        self, columns: list = None, just_schema: bool = False, **kwargs
    ) -> pyarrow.Table:
        blob_names = self.partition_scheme.get_blobs_in_partition(
            start_date=self.start_date,
            end_date=self.end_date,
            blob_list_getter=self.get_list_of_blob_names,
            prefix=self.dataset,
        )

        # Check if the first blob was cached earlier
        if self.cached_first_blob is not None:
            yield self.cached_first_blob  # Use cached blob
            blob_names = blob_names[1:]  # Skip first blob
        self.cached_first_blob = None

        for blob_name in blob_names:
            try:
                decoder = get_decoder(blob_name)
                blob_bytes = self.read_blob(blob_name=blob_name, statistics=self.statistics)
                try:
                    decoded = decoder(blob_bytes, projection=columns, just_schema=just_schema)
                except Exception as err:
                    raise DataError(f"Unable to read file {blob_name} ({err})") from err
                if not just_schema:
                    num_rows, num_columns, decoded = decoded
                    self.statistics.rows_seen += num_rows
                yield decoded
            except UnsupportedFileTypeError:
                pass

    def get_dataset_schema(self) -> RelationSchema:
        if self.schema:
            return self.schema

        # Read first blob for schema inference and cache it
        self.schema = next(self.read_dataset(just_schema=True), None)

        if self.schema is None:
            raise DatasetNotFoundError(dataset=self.dataset)

        return self.schema

    async def async_read_blob(self, *, blob_name, pool, statistics, **kwargs):
        try:
            bucket, object_path, name, extension = paths.get_parts(blob_name)
            # DEBUG: log ("READ   ", name)
            stream = self.minio.get_object(bucket, object_path + "/" + name + extension)
            data = stream.read()

            ref = await pool.commit(data)
            while ref is None:
                statistics.stalls_writing_to_read_buffer += 1
                await asyncio.sleep(0.1)
                ref = await pool.commit(data)
            statistics.bytes_read += len(data)
            return ref
        finally:
            stream.close()

    def read_blob(self, *, blob_name, **kwargs):
        try:
            bucket, object_path, name, extension = paths.get_parts(blob_name)
            stream = self.minio.get_object(bucket, object_path + "/" + name + extension)
            content = stream.read()
            self.statistics.bytes_read += len(content)
            return content
        finally:
            stream.close()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implement conditions which are essentially unary statements, usually IS statements.

This are executed as functions on arrays rather than functions on elements in arrays.
"""

import numpy
import pyarrow


def _is_null(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is null (NaN, NaT, etc.).

    Parameters:
        values: numpy.ndarray
            1D array of various types.

    Returns:
        numpy.ndarray: 1D array of booleans serving as a mask.
    """
    if isinstance(values, pyarrow.Array):
        from pyarrow import compute

        return compute.is_null(values)
    if values.dtype.kind in ("f", "b", "i"):  # float, bool, int
        return numpy.isnan(values)
    elif values.dtype.kind == "M":  # datetime64
        return numpy.isnat(values)
    elif values.dtype.kind in ("S", "O", "U"):  # string or object
        return numpy.vectorize(lambda x: x is None)(values)
    else:
        raise TypeError(
            f"Unsupported type for null comparison: {values.dtype} ({values.dtype.kind})"
        )


def _is_not_null(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is not null (NaN).

    Parameters:
        values: numpy.ndarray
            1D array of boolean and/or null values.

    Returns:
        numpy.ndarray: 1D array of booleans serving as a mask.
    """
    return numpy.logical_not(_is_null(values))


def _is_true(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is True.

    Parameters:
        values: np.ndarray
            1D array of boolean and/or null values.

    Returns:
        np.ndarray: 1D array of booleans serving as a mask.
    """
    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(False)
    return values == True


def _is_false(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is False.

    Parameters:
        values: np.ndarray
            1D array of boolean and/or null values.

    Returns:
        np.ndarray: 1D array of booleans serving as a mask.
    """
    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(False)
    return values == False


def _is_not_true(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is not True.

    Parameters:
        values: np.ndarray
            1D array of boolean and/or null values.

    Returns:
        np.ndarray: 1D array of booleans serving as a mask.
    """
    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(False)
    return values != True


def _is_not_false(values: numpy.ndarray) -> numpy.ndarray:
    """
    Returns a boolean mask where True indicates that the corresponding element in values is not False.

    Parameters:
        values: np.ndarray
            1D array of boolean and/or null values.

    Returns:
        np.ndarray: 1D array of booleans serving as a mask.
    """
    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(False)
    return values != False


UNARY_OPERATIONS = {
    "IsNull": _is_null,
    "IsNotFalse": _is_not_false,
    "IsNotNull": _is_not_null,
    "IsNotTrue": _is_not_true,
    "IsTrue": _is_true,
    "IsFalse": _is_false,
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Inner Join Node

This is a SQL Query Execution Plan Node.

PyArrow has a good LEFT JOIN implementation, but it errors when the
relations being joined contain STRUCT or ARRAY columns, this is true
for all of the JOIN types, however we've only written our own INNER
and LEFT JOINs.

It is comparible performance to the PyArrow INNER JOIN, in benchmarks
sometimes native is faster, sometimes PyArrow is faster. Generally
PyArrow is more forgiving when the relations are the "wrong" way around
(unoptimized order) but native is faster for well-ordered relations, as
we intend to take steps to help ensure relations are well-ordered, this
should work in our favour.

This is a hash join, this is completely rewritten from the earlier
pyarrow_ops implementation which was a variation of a sort-merge join.
"""

import time
from typing import Generator

import pyarrow

from opteryx.compiled.structures.hash_table import hash_join_map
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType
from opteryx.utils.arrow import align_tables


def inner_join_with_preprocessed_left_side(left_relation, right_relation, join_columns, hash_table):
    """
    Perform an INNER JOIN using a preprocessed hash table from the left relation.

    Parameters:
        left_relation: The preprocessed left pyarrow.Table.
        right_relation: The right pyarrow.Table to join.
        join_columns: A list of column names to join on.
        hash_table: The preprocessed hash table from the left table.

    Returns:
        A tuple containing lists of matching row indices from the left and right relations.
    """
    from collections import deque

    left_indexes = deque()
    right_indexes = deque()

    right_hash = hash_join_map(right_relation, join_columns)

    for h, right_rows in right_hash.hash_table.items():
        left_rows = hash_table.get(h)
        if left_rows:
            for l in left_rows:
                for r in right_rows:
                    left_indexes.append(l)
                    right_indexes.append(r)

    return align_tables(right_relation, left_relation, list(right_indexes), list(left_indexes))


class InnerJoinNode(BasePlanNode):
    operator_type = OperatorType.PASSTHRU

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._join_type = config["type"]
        self._on = config.get("on")
        self._using = config.get("using")

        self._left_columns = config.get("left_columns")
        self._left_relation = config.get("left_relation_names")

        self._right_columns = config.get("right_columns")
        self._right_relation = config.get("right_relation_names")

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    @property
    def name(self):  # pragma: no cover
        return "Inner Join"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Generator:
        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        left_relation = pyarrow.concat_tables(left_node.execute(), promote_options="none")
        # in place until #1295 resolved
        if self._left_columns[0] not in left_relation.column_names:
            self._right_columns, self._left_columns = (
                self._left_columns,
                self._right_columns,
            )

        start = time.monotonic_ns()
        left_hash = hash_join_map(left_relation, self._left_columns)

        self.statistics.time_inner_join += time.monotonic_ns() - start
        for morsel in right_node.execute():
            start = time.monotonic_ns()
            # do the join
            new_morsel = inner_join_with_preprocessed_left_side(
                left_relation=left_relation,
                right_relation=morsel,
                join_columns=self._right_columns,
                hash_table=left_hash,
            )
            self.statistics.time_inner_join += time.monotonic_ns() - start
            yield new_morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The CQL Connector downloads data from remote servers and converts them
to pyarrow tables so they can be processed as per any other data source.

CQL is Cassandra Query Language, it looks at lot like SQL.
"""

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple

import pyarrow
from orso import DataFrame
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import PYTHON_TO_ORSO_MAP
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import INITIAL_CHUNK_SIZE
from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnmetRequirementError
from opteryx.managers.expression import Node
from opteryx.managers.expression import NodeType
from opteryx.third_party.query_builder import Query


def _handle_operand(operand: Node, parameters: list) -> Tuple[Any, list]:
    if operand.node_type == NodeType.IDENTIFIER:
        return f'"{operand.source_column}"', parameters

    literal = operand.value
    if hasattr(literal, "item"):
        literal = literal.item()

    parameters.append(literal)
    return "?", parameters


class CqlConnector(BaseConnector, PredicatePushable):
    __mode__ = "Cql"
    __type__ = "CQL"

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        # "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    OPS_XLAT: Dict[str, str] = {
        "Eq": ":left = :right",
        # "NotEq": ":left not in (:right)",
        "Gt": ":left > :right",
        "GtEq": ":left >= :right",
        "Lt": ":left < :right",
        "LtEq": ":left <= :right",
        "Like": ":left LIKE :right",
        "NotLike": "NOT (:left LIKE :right)",
    }

    def __init__(
        self,
        *args,
        nodes: List[str] = None,
        username: str = None,
        password: str = None,
        cluster=None,
        **kwargs,
    ):
        BaseConnector.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)

        try:
            from cassandra.auth import PlainTextAuthProvider
            from cassandra.cluster import Cluster
        except ImportError as err:  # pragma: nocover
            raise MissingDependencyError(err.name) from err

        if cluster is None and username is None:  # pragma: no cover
            raise UnmetRequirementError(
                "CQL Connections require either a Cassandra Cluster in the 'cluster' parameter, or a Cassandra cluster nodes in the 'nodes' parameter, usually with the 'username' and 'password' parameters set."
            )

        if username:
            auth_provider = PlainTextAuthProvider(username=username, password=password)
            self.cluster = Cluster(nodes, auth_provider=auth_provider)
        else:
            self.cluster = cluster

        self.single_column = None

    def read_dataset(  # type:ignore
        self,
        *,
        columns: list = None,
        predicates: list = None,
        chunk_size: int = INITIAL_CHUNK_SIZE,  # type:ignore
    ) -> Generator[pyarrow.Table, None, None]:  # type:ignore
        self.chunk_size = chunk_size

        result_schema = self.schema

        query_builder = Query().FROM(self.dataset)

        # Update the SQL and the target morsel schema if we've pushed a projection
        if columns:
            column_names = [f'"{col.source_column}"' for col in columns]
            query_builder.add("SELECT", *column_names)
            result_schema.columns = [  # type:ignore
                col
                for col in self.schema.columns  # type:ignore
                if f'"{col.name}"' in column_names  # type:ignore
            ]
        else:
            query_builder.add("SELECT", f'"{self.single_column.name}"')  # type:ignore
            self.schema.columns = [self.single_column]  # type:ignore

        # Update SQL if we've pushed predicates
        parameters: list = []
        for predicate in predicates:
            left_operand = predicate.left
            right_operand = predicate.right
            operator = self.OPS_XLAT[predicate.value]

            left_value, parameters = _handle_operand(left_operand, parameters)
            right_value, parameters = _handle_operand(right_operand, parameters)

            operator = operator.replace(":left", left_value)
            operator = operator.replace(":right", right_value)

            query_builder.WHERE(operator)

        session = self.cluster.connect()
        # DEBUG: log ("READ DATASET\n", str(query_builder))
        # DEBUG: log ("PARAMETERS\n", parameters)
        # Execution Options allows us to handle datasets larger than memory
        statement = session.prepare(str(query_builder) + " ALLOW FILTERING;")
        results = session.execute(statement, parameters=parameters)

        at_least_once = False
        for morsel in self.chunk_dictset(
            (doc._asdict() for doc in results),
            initial_chunk_size=chunk_size,
        ):
            at_least_once = True
            yield morsel

        if not at_least_once:
            yield DataFrame(schema=result_schema).arrow()

    def get_dataset_schema(self) -> RelationSchema:
        if self.schema:
            return self.schema

        # get the schema from the dataset

        session = self.cluster.connect()
        query = Query().SELECT("*").FROM(self.dataset).LIMIT("25")
        statement = session.prepare(str(query))
        # DEBUG: log ("READ 1 ROWS\n", str(query))
        results = session.execute(statement)[0]
        columns = list(results._asdict().keys())
        # DEBUG: log ("COLUMNS:", columns)
        self.schema = RelationSchema(
            name=self.dataset,
            columns=[
                FlatColumn(
                    name=column,
                    type=PYTHON_TO_ORSO_MAP.get(
                        type(getattr(results, column)), OrsoTypes._MISSING_TYPE
                    ),
                )
                for column in columns
            ],
        )

        self.single_column = self.schema.columns[0]  # type:ignore

        return self.schema



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Heap Sort Node

This is a SQL Query Execution Plan Node.

This node orders a dataset, note that Heap Sort in this instance isn't the heap sort
algorithm, it is an approach where a heap of n items (the limit) is maintained as the
data passes through the operator. Because we are working with chunks, we build small
batches which we order and then discard the excess items.

This is faster, particularly when working with large datasets even though we're now
sorting smaller chunks over and over again.
"""

import time
from dataclasses import dataclass
from typing import Generator

import pyarrow
from pyarrow import concat_tables

from opteryx.exceptions import ColumnNotFoundError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType
from opteryx.operators.base_plan_node import BasePlanDataObject


@dataclass
class HeapSortDataObject(BasePlanDataObject):
    order_by: list = None
    limit: int = -1


class HeapSortNode(BasePlanNode):
    operator_type = OperatorType.BLOCKING

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.order_by = config.get("order_by", [])
        self.limit: int = config.get("limit", -1)

        self.do = HeapSortDataObject(order_by=self.order_by, limit=self.limit)

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    @property
    def config(self):  # pragma: no cover
        return f"LIMIT = {self.limit} ORDER = " + ", ".join(
            f"{i[0].value} {i[1][0:3].upper()}" for i in self.order_by
        )

    @property
    def name(self):  # pragma: no cover
        return "Heap Sort"

    def execute(self) -> Generator[pyarrow.Table, None, None]:  # pragma: no cover
        table = None
        morsels = self._producers[0]  # type:ignore

        mapped_order = []

        for column, direction in self.order_by:
            try:
                mapped_order.append(
                    (
                        column.schema_column.identity,
                        direction,
                    )
                )
            except ColumnNotFoundError as cnfe:
                raise ColumnNotFoundError(
                    f"`ORDER BY` must reference columns as they appear in the `SELECT` clause. {cnfe}"
                )

        for morsel in morsels.execute():
            start_time = time.time_ns()

            if morsel.num_rows > self.limit:
                # not much point doing this here if we're not eliminating rows
                morsel = morsel.sort_by(mapped_order)
                morsel = morsel.slice(offset=0, length=self.limit)

            if table:
                # Concatenate the current morsel with the previously accumulated table
                morsel = concat_tables([morsel, table], promote_options="permissive")

            # Sort and slice the concatenated table to maintain the limit
            morsel = morsel.sort_by(mapped_order)
            morsel = morsel.slice(offset=0, length=self.limit)
            table = morsel

            self.statistics.time_heap_sorting += time.time_ns() - start_time

        yield table



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Distinct Pushdown

Type: Heuristic
Goal: Reduce Rows

This is a very specific rule, on a CROSS JOIN UNNEST, if the result
is the only column in a DISTINCT clause, we push the DISTINCT into
the JOIN.

We've written as a Optimization rule rather than in the JOIN code
as it is expected other instances of pushing DISTINCT may be found.

Order:
    This plan must run after the Projection Pushdown
"""

from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext

"""
Aggregations we can push the DISTINCT past
"""


class DistinctPushdownStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if (node.node_type == LogicalPlanStepType.Distinct) and node.on is None:
            node.nid = context.node_id
            context.collected_distincts.append(node)
            return context

        if (
            node.node_type == LogicalPlanStepType.Join
            and context.collected_distincts
            and node.type == "cross join"
            and node.unnest_target is not None
            and node.pre_update_columns == node.unnest_target.identity
        ):
            node.distinct = True
            context.optimized_plan[context.node_id] = node
            for distict_node in context.collected_distincts:
                context.optimized_plan.remove_node(distict_node.nid, heal=True)
            context.collected_distincts.clear()
            return context

        if node.node_type in (
            LogicalPlanStepType.Join,
            LogicalPlanStepType.Scan,
            LogicalPlanStepType.AggregateAndGroup,
            LogicalPlanStepType.Aggregate,
            LogicalPlanStepType.Subquery,
        ):
            # we don't push past here
            context.collected_distincts.clear()

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Read Node

This is the SQL Query Execution Plan Node responsible for the reading of data.

It wraps different internal readers (e.g. GCP Blob reader, SQL Reader),
normalizes the data into the format for internal processing.
"""

import time
from typing import Generator

import orjson
import pyarrow
from orso.schema import RelationSchema
from orso.schema import convert_orso_schema_to_arrow_schema

from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType


def struct_to_jsonb(table: pyarrow.Table) -> pyarrow.Table:
    """
    Converts any STRUCT columns in a PyArrow Table to JSON strings and replaces them
    in the same column position.

    Parameters:
        table (pa.Table): The PyArrow Table to process.

    Returns:
        pa.Table: A new PyArrow Table with STRUCT columns converted to JSON strings.
    """
    for i in range(table.num_columns):
        field = table.schema.field(i)

        # Check if the column is a STRUCT
        if pyarrow.types.is_struct(field.type):
            # Convert each row in the STRUCT column to a JSON string
            json_strings = [
                orjson.dumps(row.as_py()) if row.is_valid else None for row in table.column(i)
            ]
            json_array = pyarrow.array(json_strings, type=pyarrow.binary())

            # Drop the original STRUCT column
            table = table.drop_columns(field.name)

            # Insert the new JSON column at the same position
            table = table.add_column(
                i, pyarrow.field(name=field.name, type=pyarrow.binary()), json_array
            )

    return table


def normalize_morsel(schema: RelationSchema, morsel: pyarrow.Table) -> pyarrow.Table:
    if len(schema.columns) == 0:
        one_column = pyarrow.array([1] * morsel.num_rows, type=pyarrow.int8())
        morsel = morsel.append_column("*", one_column)
        return morsel.select(["*"])

    # rename columns for internal use
    target_column_names = []
    # columns in the data but not in the schema, droppable
    droppable_columns = []

    # Find which columns to drop and which columns we already have
    for i, column in enumerate(morsel.column_names):
        column_name = schema.find_column(column)
        if column_name is None:
            droppable_columns.append(i)
        else:
            target_column_names.append(str(column_name))

    # Remove from the end otherwise we'll remove the wrong columns after we've removed one
    droppable_columns.reverse()
    for droppable in droppable_columns:
        morsel = morsel.remove_column(droppable)

    # remane columns to the internal names (identities)
    morsel = morsel.rename_columns(target_column_names)

    # add columns we don't have, populate with nulls but try to get the correct type
    for column in schema.columns:
        if column.identity not in target_column_names:
            null_column = pyarrow.array([None] * morsel.num_rows)
            field = pyarrow.field(name=column.identity, type=column.arrow_field.type)
            morsel = morsel.append_column(field, null_column)

    # ensure the columns are in the right order
    return morsel.select([col.identity for col in schema.columns])


def merge_schemas(
    hypothetical_schema: RelationSchema, observed_schema: pyarrow.Schema
) -> pyarrow.schema:
    """
    Using the hypothetical schema as the base, replace with fields from the observed schema
    which are a Decimal type.
    """
    # convert the Orso schema to an Arrow schema
    hypothetical_arrow_schema = convert_orso_schema_to_arrow_schema(hypothetical_schema, True)

    # Convert the hypothetical schema to a dictionary for easy modification
    schema_dict = {field.name: field for field in hypothetical_arrow_schema}

    # Iterate through fields in the observed schema
    for observed_field in observed_schema:
        # Check if the field is of type Decimal or List/Array
        if pyarrow.types.is_decimal(observed_field.type) or pyarrow.types.is_list(
            observed_field.type
        ):
            # Replace or add the field to the schema dictionary
            schema_dict[observed_field.name] = observed_field

    # Create a new schema from the updated dictionary of fields
    merged_schema = pyarrow.schema(list(schema_dict.values()))

    return merged_schema


class ReaderNode(BasePlanNode):
    operator_type = OperatorType.PRODUCER

    def __init__(self, properties: QueryProperties, **parameters):
        super().__init__(properties=properties, **parameters)
        self.start_date = parameters.get("start_date")
        self.end_date = parameters.get("end_date")
        self.hints = parameters.get("hints", [])
        self.columns = parameters.get("columns", [])
        self.predicates = parameters.get("predicates", [])

        self.connector = parameters.get("connector")
        self.schema = parameters.get("schema")

        if len(self.hints) != 0:
            self.statistics.add_message("All HINTS are currently ignored")

    def to_dict(self) -> dict:
        return {
            "identity": f"read-{self.identity}",
            "opterator": "ReadNode",
            "schema": self.columns,
            "projection": self.columns,
            "filters": self.predicates,
        }

    @classmethod
    def from_dict(cls, dic: dict) -> "BasePlanNode":
        raise NotImplementedError()

    @property
    def name(self):  # pragma: no cover
        """friendly name for this step"""
        return "Read"

    @property
    def config(self):
        """Additional details for this step"""
        date_range = ""
        if self.parameters.get("start_date") == self.parameters.get("end_date"):
            if self.parameters.get("start_date") is not None:
                date_range = f" FOR '{self.parameters.get('start_date')}'"
        else:
            date_range = (
                f" FOR '{self.parameters.get('start_date')}' TO '{self.parameters.get('end_date')}'"
            )
        return (
            f"{self.connector.__type__} "
            f"({self.parameters.get('relation')}"
            f"{' AS ' + self.parameters.get('alias') if self.parameters.get('alias') else ''}"
            f"{date_range}"
            f"{' WITH(' + ','.join(self.parameters.get('hints')) + ')' if self.parameters.get('hints') else ''})"
        )

    def execute(self) -> Generator:
        """Perform this step, time how long is spent doing work"""
        morsel = None
        orso_schema = self.schema
        orso_schema_cols = []
        for col in orso_schema.columns:
            if col.identity in [c.schema_column.identity for c in self.columns]:
                orso_schema_cols.append(col)
        orso_schema.columns = orso_schema_cols
        arrow_schema = None
        start_clock = time.monotonic_ns()
        reader = self.connector.read_dataset(columns=self.columns, predicates=self.predicates)
        for morsel in reader:
            # try to make each morsel have the same schema
            morsel = struct_to_jsonb(morsel)
            morsel = normalize_morsel(orso_schema, morsel)
            if arrow_schema is None:
                arrow_schema = merge_schemas(self.schema, morsel.schema)
            if arrow_schema.names:
                morsel = morsel.cast(arrow_schema)

            self.statistics.time_reading_blobs += time.monotonic_ns() - start_clock
            self.statistics.blobs_read += 1
            self.statistics.rows_read += morsel.num_rows
            self.statistics.bytes_processed += morsel.nbytes
            yield morsel
            start_clock = time.monotonic_ns()
        if morsel:
            self.statistics.columns_read += morsel.num_columns



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some operators can be fused to be faster.

'Fused' opertors are when physical operations perform multiple logical operations.

Initially we fused Limit and Order operators, this allows us to use a heap sort
algorithm (basically we dicard records we know aren't going to be kept early).

Note that predicate and projection pushdowns may also fuse operators. Most commonly
we fuse the READ operator with SELECTION and PROJECTION operators, we also push into
JOINs, this is sometimes as part of the join condition, but we also push SELECTIONs
into joins.
"""

from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext


class OperatorFusionStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Order:
            edges = context.optimized_plan.outgoing_edges(context.node_id)
            if len(edges) == 1:
                next_node_id = edges[0][1]
                next_node = context.optimized_plan[next_node_id]
                if next_node.node_type == LogicalPlanStepType.Limit and not next_node.offset:
                    new_node = LogicalPlanNode(node_type=LogicalPlanStepType.HeapSort)
                    new_node.limit = next_node.limit
                    new_node.order_by = node.order_by
                    context.optimized_plan[next_node_id] = new_node
                    context.optimized_plan.remove_node(context.node_id, heal=True)

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        return plan



import re


def tokenize_string(string):
    # Define the regular expression to match tokens
    token_pattern = re.compile(r"[(),;]|[\r\n]+|[^\s(),;\r\n]+")

    # Use the regular expression to find all tokens in the string
    tokens = token_pattern.findall(string)

    # Remove any leading or trailing whitespace from the tokens
    tokens = ["\n" if token == "\n" else token.strip() for token in tokens]

    return tokens


def format_sql(sql):  # pragma: no cover
    """
    Adds colorization to SQL statements to make it easier to find keywords and literals
    """

    def color_comments(string):  # pragma: no cover
        pattern = r"(\"[^\"]\"|\'[^\']\'|\`[^\`]\`)|(/\*[^\*/]*\*/|--[^\r\n]*$)"
        regex = re.compile(pattern, re.MULTILINE | re.DOTALL)
        ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")

        def _replacer(match):
            if match.group(2) is not None:
                return f"\033[38;2;98;114;164m\033[3m{ansi_escape.sub('', match.group(2))}\033[0m"
            return match.group(1)

        return regex.sub(_replacer, string)

    if hasattr(sql, "decode"):
        sql = sql.decode()
    words = tokenize_string(sql)

    formatted_sql = ""
    in_string_literal = False

    for i, word in enumerate(words):
        if not in_string_literal and word.startswith("'"):
            formatted_sql += "\033[38;2;255;171;82m" + word
            if word.endswith("'"):
                in_string_literal = False
                formatted_sql += "\033[0m "
            else:
                in_string_literal = True
        elif in_string_literal:
            formatted_sql += " " + word
            if word.endswith("'"):
                in_string_literal = False
                formatted_sql += "\033[0m "
        elif word in ("(", ")", ",", ";"):
            formatted_sql += "\033[38;5;102m" + word + "\033[0m "
        elif word.upper() in {
            "ANALYZE",
            "ANTI",
            "AS",
            "BY",
            "CROSS",
            "DATES",
            "DISTINCT",
            "EXPLAIN",
            "FOR",
            "FROM",
            "FULL",
            "GROUP",
            "HAVING",
            "INNER",
            "INTERVAL",
            "JOIN",
            "LEFT",
            "LIMIT",
            "OFFSET",
            "ON",
            "ORDER",
            "OUTER",
            "RIGHT",
            "SELECT",
            "SET",
            "SHOW",
            "SINCE",
            "TODAY",
            "UNION",
            "USE",
            "USING",
            "WHERE",
            "WITH",
            "YESTERDAY",
        }:
            formatted_sql += "\033[38;2;139;233;253m" + word.upper() + "\033[0m "
        elif word.upper() in ("TRUE", "FALSE", "NULL", "DAY", "MONTH", "MINUTE"):
            formatted_sql += "\033[38;2;255;184;108m" + word.upper() + "\033[0m "
        elif (i + 1) < len(words) and words[i + 1] == "(":
            formatted_sql += "\033[38;2;80;250;123m" + word.upper() + "\033[0m"
        elif word.upper() in (
            "=",
            ">=",
            "<=",
            "!=",
            "<",
            ">",
            "<>",
            "-",
            "+",
            "LIKE",
            "ILIKE",
            "RLIKE",
            "NOT",
            "AND",
            "OR",
            "IN",
            "SIMILAR",
            "TO",
            "BETWEEN",
            "IS",
        ):
            formatted_sql += "\033[38;2;189;147;249m" + word.upper() + "\033[0m "
        elif word.replace(".", "", 1).isdigit():
            formatted_sql += "\033[38;2;255;184;108m" + word + "\033[0m "
        elif word == "\n":
            formatted_sql = formatted_sql.strip() + "\n"
        else:
            formatted_sql += word + " "

    formatted_sql += "\033[0m"

    formatted_sql = formatted_sql.replace(" \033[38;5;102m(", "\033[38;5;102m(")
    formatted_sql = formatted_sql.replace("(\033[0m ", "(\033[0m")
    formatted_sql = formatted_sql.replace(" \033[38;5;102m)", "\033[38;5;102m)")
    formatted_sql = formatted_sql.replace(" \033[38;5;102m,", "\033[38;5;102m,")
    formatted_sql = formatted_sql.replace(" \033[38;5;102m;", "\033[38;5;102m;")

    return color_comments(formatted_sql).strip()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Owner meanings:
    SERVER - can only be set at the server level at start up (set in config)
    INTERNAL - the system can update this as it runs (defaulted in config)
    USER - the user can update this value (defaulted in config)

For variables we're creating and naming, use sensible defaults and if it's a
feature flag, name the variable for the state the user probably doesn't want -
e.g. disable_optimizer (default to False)
"""

from enum import Enum
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Type

from orso.types import OrsoTypes

from opteryx import config
from opteryx.__version__ import __version__
from opteryx.constants.character_set import CharacterSet
from opteryx.constants.character_set import Collation
from opteryx.exceptions import PermissionsError
from opteryx.exceptions import VariableNotFoundError


class VariableOwner(int, Enum):
    # Manually assign numbers because USER < INTERNAL < SERVER
    SERVER = 30
    INTERNAL = 20
    USER = 10


VariableSchema = Tuple[Type, Any, VariableOwner]

# fmt: off
SYSTEM_VARIABLES_DEFAULTS: Dict[str, VariableSchema] = {
    # name: (type, default, owner, description)

    # These are the MySQL set of variables - we don't use all of them but have them for compatibility
    "auto_increment_increment": (OrsoTypes.INTEGER, 1, VariableOwner.INTERNAL),
    "autocommit": (OrsoTypes.BOOLEAN, True, VariableOwner.SERVER),
    "character_set_client": (OrsoTypes.VARCHAR, CharacterSet.utf8mb4.name, VariableOwner.SERVER),
    "character_set_connection": (OrsoTypes.VARCHAR, CharacterSet.utf8mb4.name, VariableOwner.SERVER),
    "character_set_database": (OrsoTypes.VARCHAR, CharacterSet.utf8mb4.name, VariableOwner.SERVER),
    "character_set_results": (OrsoTypes.VARCHAR, CharacterSet.utf8mb4.name, VariableOwner.SERVER),
    "character_set_server": (OrsoTypes.VARCHAR, CharacterSet.utf8mb4.name, VariableOwner.SERVER),
    "collation_connection": (OrsoTypes.VARCHAR, Collation.utf8mb4_general_ci.name, VariableOwner.SERVER),
    "collation_database": (OrsoTypes.VARCHAR, Collation.utf8mb4_general_ci.name, VariableOwner.SERVER),
    "collation_server": (OrsoTypes.VARCHAR, Collation.utf8mb4_general_ci.name, VariableOwner.SERVER),
    "external_user": (OrsoTypes.VARCHAR, "", VariableOwner.INTERNAL),
    "init_connect": (OrsoTypes.VARCHAR, "", VariableOwner.SERVER),
    "interactive_timeout": (OrsoTypes.INTEGER, 28800, VariableOwner.SERVER),
    "license": (OrsoTypes.VARCHAR, "MIT", VariableOwner.SERVER),
    "lower_case_table_names": (OrsoTypes.INTEGER, 0, VariableOwner.SERVER),
    "max_allowed_packet": (OrsoTypes.INTEGER, 67108864, VariableOwner.SERVER),
    "max_execution_time": (OrsoTypes.INTEGER, 0, VariableOwner.SERVER),
    "net_buffer_length": (OrsoTypes.INTEGER, 16384, VariableOwner.SERVER),
    "net_write_timeout": (OrsoTypes.INTEGER, 28800, VariableOwner.SERVER),
    "performance_schema": (OrsoTypes.BOOLEAN, False, VariableOwner.SERVER),
    "sql_auto_is_null": (OrsoTypes.BOOLEAN, False, VariableOwner.SERVER),
    "sql_mode": (OrsoTypes.VARCHAR, "ANSI", VariableOwner.SERVER),
    "sql_select_limit": (OrsoTypes.INTEGER, None, VariableOwner.SERVER),
    "system_time_zone": (OrsoTypes.VARCHAR, "UTC", VariableOwner.SERVER),
    "time_zone": (OrsoTypes.VARCHAR, "UTC", VariableOwner.SERVER),
    "transaction_read_only": (OrsoTypes.BOOLEAN, False, VariableOwner.SERVER),
    "transaction_isolation": (OrsoTypes.VARCHAR, "READ-COMMITTED", VariableOwner.SERVER),
    "version": (OrsoTypes.VARCHAR, __version__, VariableOwner.SERVER),
    "version_comment": (OrsoTypes.VARCHAR, "mesos", VariableOwner.SERVER),
    "wait_timeout": (OrsoTypes.INTEGER, 28800, VariableOwner.SERVER),
    "event_scheduler": (OrsoTypes.VARCHAR, "OFF", VariableOwner.SERVER),
    "default_storage_engine": (OrsoTypes.VARCHAR, "opteryx", VariableOwner.SERVER),
    "default_tmp_storage_engine": (OrsoTypes.VARCHAR, "opteryx", VariableOwner.SERVER),

    # these are Opteryx specific variables
    "max_cache_evictions_per_query": (OrsoTypes.INTEGER, config.MAX_CACHE_EVICTIONS_PER_QUERY, VariableOwner.USER),
    "max_cacheable_item_size": (OrsoTypes.INTEGER, config.MAX_CACHEABLE_ITEM_SIZE, VariableOwner.SERVER),
    "max_local_buffer_capacity": (OrsoTypes.INTEGER, config.MAX_LOCAL_BUFFER_CAPACITY, VariableOwner.SERVER),
    "max_read_buffer_capacity": (OrsoTypes.INTEGER, config.MAX_READ_BUFFER_CAPACITY, VariableOwner.SERVER),
    "disable_optimizer": (OrsoTypes.BOOLEAN, config.DISABLE_OPTIMIZER, VariableOwner.USER),
    "disable_high_priority": (OrsoTypes.BOOLEAN, config.DISABLE_HIGH_PRIORITY, VariableOwner.SERVER),
    "concurrent_reads": (OrsoTypes.BOOLEAN, config.CONCURRENT_READS, VariableOwner.SERVER),
    "user_memberships": (OrsoTypes.ARRAY, [], VariableOwner.INTERNAL),
    "morsel_size": (OrsoTypes.INTEGER, config.MORSEL_SIZE, VariableOwner.SERVER),
}
# fmt: on


class SystemVariablesContainer:
    def __init__(self, owner: VariableOwner = VariableOwner.USER):
        self._variables = SYSTEM_VARIABLES_DEFAULTS.copy()
        self._owner = owner

    def __getitem__(self, key: str) -> Any:
        if key not in self._variables:
            raise VariableNotFoundError(key)
        return self._variables[key][1]

    def __setitem__(self, key: str, value: Any) -> None:
        if key[0] == "@":
            variable_type = value.type
            owner = VariableOwner.USER
        else:
            if key not in self._variables:
                from opteryx.utils import suggest_alternative

                suggestion = suggest_alternative(key, list(self._variables.keys()))

                raise VariableNotFoundError(variable=key, suggestion=suggestion)
            variable_type, _, owner = self._variables[key]
            if owner > self._owner:
                raise PermissionsError(f"User does not have permission to set variable `{key}`")
            if variable_type != value.type:
                raise ValueError(f"Invalid type for `{key}`, {variable_type} expected.")

        self._variables[key] = (variable_type, value.value, owner)

    def details(self, key: str) -> VariableSchema:
        if key not in self._variables:
            raise VariableNotFoundError(key)
        return self._variables[key]

    def __contains__(self, key: str) -> bool:
        return key in self._variables

    def __iter__(self):
        return iter(self._variables)

    def __len__(self):
        return len(self._variables)

    def snapshot(self, owner: VariableOwner = VariableOwner.USER) -> "SystemVariablesContainer":
        return SystemVariablesContainer(owner)

    def as_column(self, key: str):
        """Return a variable as a CONSTANT column"""
        from orso.schema import ConstantColumn

        # system variables aren't stored with the @@
        variable = self._variables[key[2:]] if key.startswith("@@") else self._variables[key]
        return ConstantColumn(name=key, type=variable[0], value=variable[1])


# load the base set
SystemVariables = SystemVariablesContainer(VariableOwner.INTERNAL)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PREDICATE REWRITER

We rewrite some conditions to a more optimal form; for example if doing a
LIKE comparison and the pattern contains no wildcards, we rewrite to be an
equals check.

Rewrite to a form which is just faster, even if it can't be pushed
"""

import re
from typing import Callable
from typing import Dict

from orso.types import OrsoTypes

from opteryx.managers.expression import ExpressionColumn
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import format_expression
from opteryx.models import Node
from opteryx.planner.binder.operator_map import determine_type
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext

IN_REWRITES = {"InList": "Eq", "NotInList": "NotEq"}
LIKE_REWRITES = {"Like": "Eq", "NotLike": "NotEq"}
LITERALS_TO_THE_RIGHT = {"Plus": "Minus", "Minus": "Plus"}


def _add_condition(existing_condition, new_condition):
    if not existing_condition:
        return new_condition
    _and = Node(node_type=NodeType.AND)
    _and.left = new_condition
    _and.right = existing_condition
    return _and


def remove_adjacent_wildcards(predicate):
    """
    Remove adjacent wildcards from LIKE/ILIKE/NotLike/NotILike conditions.

    This optimization removes redundant wildcards from LIKE patterns to help
    improve matching with subsequent optimizations
    """
    if "%%" in predicate.right.value:
        predicate.right.value = re.sub(r"%+", "%", predicate.right.value)
    return predicate


def rewrite_to_starts_with(predicate):
    """
    Rewrite LIKE/ILIKE conditions with a single trailing wildcard to STARTS_WITH function.

    This optimization converts patterns like 'abc%' to a STARTS_WITH function, which can be
    more efficiently processed by the underlying engine compared to a generic LIKE pattern.
    """
    ignore_case = predicate.value == "ILike"
    predicate.right.value = predicate.right.value[:-1]
    predicate.node_type = NodeType.FUNCTION
    predicate.value = "STARTS_WITH"
    predicate.parameters = [
        predicate.left,
        predicate.right,
        Node(node_type=NodeType.LITERAL, type=OrsoTypes.BOOLEAN, value=ignore_case),
    ]
    return predicate


def rewrite_to_ends_with(predicate):
    """
    Rewrite LIKE/ILIKE conditions with a single leading wildcard to ENDS_WITH function.

    This optimization converts patterns like '%abc' to an ENDS_WITH function, which can be
    more efficiently processed by the underlying engine compared to a generic LIKE pattern.
    """
    ignore_case = predicate.value == "ILike"
    predicate.right.value = predicate.right.value[1:]
    predicate.node_type = NodeType.FUNCTION
    predicate.value = "ENDS_WITH"
    predicate.parameters = [
        predicate.left,
        predicate.right,
        Node(node_type=NodeType.LITERAL, type=OrsoTypes.BOOLEAN, value=ignore_case),
    ]
    return predicate


def rewrite_to_search(predicate):
    """
    Rewrite LIKE/ILIKE conditions with leading and trailing wildcards to SEARCH function.

    This optimization converts patterns like '%abc%' to a SEARCH function, which can be
    more efficiently processed by the underlying engine compared to a generic LIKE pattern.
    """
    ignore_case = predicate.value == "ILike"
    predicate.right.value = predicate.right.value[1:-1]
    predicate.node_type = NodeType.FUNCTION
    predicate.value = "SEARCH"
    predicate.parameters = [
        predicate.left,
        predicate.right,
        Node(node_type=NodeType.LITERAL, type=OrsoTypes.BOOLEAN, value=ignore_case),
    ]
    return predicate


def rewrite_in_to_eq(predicate):
    """
    Rewrite IN conditions with a single value to equality conditions.

    If the IN condition contains only one value, it is equivalent to an equality check.
    This optimization replaces the IN condition with a faster equality check.
    """
    predicate.value = IN_REWRITES[predicate.value]
    predicate.right.value = predicate.right.value.pop()
    predicate.right.type = predicate.right.sub_type or OrsoTypes.VARCHAR
    predicate.right.sub_type = None
    return predicate


def reorder_interval_calc(predicate):
    """
    rewrite:
        end - start > interval => start + interval > end

    This is because comparing a Date with a Date is faster than
    comparing in Interval with an Interval.
    """
    date_start = predicate.left.right
    date_end = predicate.left.left
    interval = predicate.right

    # Check if the operation is date - date
    if predicate.left.value == "Minus":
        # Create a new binary operator node for date + interval
        new_binary_op = Node(
            node_type=NodeType.BINARY_OPERATOR,
            value="Plus",
            left=date_start,
            right=interval,
        )
        binary_op_column_name = format_expression(new_binary_op, True)
        new_binary_op.schema_column = ExpressionColumn(
            name=binary_op_column_name, type=OrsoTypes.TIMESTAMP
        )

        # Create a new comparison operator node for date > date
        predicate.node_type = NodeType.COMPARISON_OPERATOR
        predicate.right = new_binary_op
        predicate.left = date_end

        predicate_column_name = format_expression(predicate, True)
        predicate.schema_column = ExpressionColumn(
            name=predicate_column_name, type=OrsoTypes.BOOLEAN
        )

        return predicate


# Define dispatcher conditions and actions
dispatcher: Dict[str, Callable] = {
    "remove_adjacent_wildcards": remove_adjacent_wildcards,
    "rewrite_to_starts_with": rewrite_to_starts_with,
    "rewrite_to_ends_with": rewrite_to_ends_with,
    "rewrite_to_search": rewrite_to_search,
    "rewrite_in_to_eq": rewrite_in_to_eq,
    "reorder_interval_calc": reorder_interval_calc,
}


# Dispatcher conditions
def _rewrite_predicate(predicate):
    if predicate.node_type not in {NodeType.BINARY_OPERATOR, NodeType.COMPARISON_OPERATOR}:
        # after rewrites, some filters aren't actually predicates
        return predicate

    if predicate.node_type in {NodeType.AND, NodeType.OR, NodeType.XOR}:
        predicate.left = _rewrite_predicate(predicate.left)
        predicate.right = _rewrite_predicate(predicate.right)

    if predicate.value in {"Like", "ILike", "NotLike", "NotILike"}:
        if "%%" in predicate.right.value:
            predicate = dispatcher["remove_adjacent_wildcards"](predicate)

    if predicate.value in {"Like", "NotLike"}:
        if "%" not in predicate.right.value and "_" not in predicate.right.value:
            predicate.value = LIKE_REWRITES[predicate.value]

    if predicate.value in {"Like", "ILike"}:
        if predicate.left.source_connector and predicate.left.source_connector.isdisjoint(
            {"Sql", "Cql"}
        ):
            if predicate.right.node_type == NodeType.LITERAL:
                if predicate.right.value[-1] == "%" and predicate.right.value.count("%") == 1:
                    return dispatcher["rewrite_to_starts_with"](predicate)
                if predicate.right.value[0] == "%" and predicate.right.value.count("%") == 1:
                    return dispatcher["rewrite_to_ends_with"](predicate)
                if (
                    predicate.right.value[0] == "%"
                    and predicate.right.value[-1] == "%"
                    and predicate.right.value.count("%") == 2
                ):
                    return dispatcher["rewrite_to_search"](predicate)

    if predicate.value == "AnyOpEq":
        if predicate.right.node_type == NodeType.LITERAL:
            predicate.value = "InList"

    if predicate.value in IN_REWRITES:
        if predicate.right.node_type == NodeType.LITERAL and len(predicate.right.value) == 1:
            return dispatcher["rewrite_in_to_eq"](predicate)

    if (
        predicate.node_type == NodeType.COMPARISON_OPERATOR
        and predicate.left.node_type == NodeType.BINARY_OPERATOR
    ):
        if (
            determine_type(predicate.left) == OrsoTypes.INTERVAL
            and determine_type(predicate.right) == OrsoTypes.INTERVAL
        ):
            predicate = dispatcher["reorder_interval_calc"](predicate)

    return predicate


class PredicateRewriteStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            node.condition = _rewrite_predicate(node.condition)
            context.optimized_plan[context.node_id] = node

        if node.node_type == LogicalPlanStepType.Project:
            new_columns = []
            for column in node.columns:
                new_column = _rewrite_predicate(column)
                new_columns.append(new_column)
            node.columns = new_columns

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This implements an interface to Memcached

If we have 10 failures in a row, stop trying to use the cache. We have some
scenarios where we assume the remote server is down and stop immediately.
"""

import os
from typing import Union

from orso.tools import single_item_cache

from opteryx.exceptions import MissingDependencyError
from opteryx.managers.kvstores import BaseKeyValueStore

MAXIMUM_CONSECUTIVE_FAILURES: int = 10


@single_item_cache
def _memcached_server(**kwargs):
    """
    Handling connecting to Memcached
    """
    # the server must be set in the environment
    memcached_config = kwargs.get("server", os.environ.get("MEMCACHED_SERVER"))
    if memcached_config is None:
        return None

    # expect either SERVER or SERVER:PORT entries
    memcached_config = memcached_config.split(":")
    if len(memcached_config) == 1:
        # the default memcached port
        memcached_config.append(11211)

    # we need the server and the port
    if len(memcached_config) != 2:
        return None

    try:
        from pymemcache.client import base
    except ImportError as err:
        raise MissingDependencyError(err.name) from err

    try:
        # wait 1 second to try to connect, it's not worthwhile as a cache if it's slow
        cache = base.Client(
            (
                memcached_config[0],
                memcached_config[1],
            ),
            connect_timeout=1,
            timeout=1,
        )
    except Exception as err:
        print("[CACHE] Unable to create remote cache", err)
        cache = None

    return cache


class MemcachedCache(BaseKeyValueStore):
    """
    Cache object
    """

    def __init__(self, **kwargs):
        """
        Parameters:
            servers: string (optional)
                Sets the memcached server and port (server:port). If not provided
                the value will be obtained from the OS environment.
        """
        self._server = _memcached_server(**kwargs)
        if self._server is None:
            self._consecutive_failures: int = MAXIMUM_CONSECUTIVE_FAILURES
        else:
            self._consecutive_failures = 0
        self.hits: int = 0
        self.misses: int = 0
        self.skips: int = 0
        self.errors: int = 0
        self.sets: int = 0
        self.touches: int = 0

    def get(self, key: bytes) -> Union[bytes, None]:
        if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
            self.skips += 1
            return None
        try:
            response = self._server.get(key)
            self._consecutive_failures = 0
            if response:
                self.hits += 1
                return bytes(response)
        except Exception as err:  # pragma: no cover
            self._consecutive_failures += 1
            if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
                import datetime

                print(
                    f"{datetime.datetime.now()} [CACHE] Disabling remote Memcached cache due to persistent errors ({err}) [GET]."
                )
            self.errors += 1
            return None

        self.misses += 1
        return None

    def set(self, key: bytes, value: bytes) -> None:
        if self._consecutive_failures < MAXIMUM_CONSECUTIVE_FAILURES:
            try:
                self._server.set(key, value)
                self.sets += 1
            except Exception as err:
                # if we fail to set, stop trying
                self._consecutive_failures = MAXIMUM_CONSECUTIVE_FAILURES
                self.errors += 1
                import datetime

                print(
                    f"{datetime.datetime.now()} [CACHE] Disabling remote Memcached cache due to persistent errors ({err}) [SET]."
                )
        else:
            self.skips += 1

    def touch(self, key: bytes):
        if self._consecutive_failures < MAXIMUM_CONSECUTIVE_FAILURES:
            try:
                self._server.touch(key)
                self.touches += 1
            except Exception as err:  # nosec
                if not err:
                    pass
                # DEBUG: log(f"Unable to 'touch' Memcache cache {err}")
                self.errors += 1
                pass

    def __del__(self):
        pass
        # DEBUG: log(f"Memcached <hits={self.hits} misses={self.misses} sets={self.sets} skips={self.skips} errors={self.errors} touches={self.touches}>")



# cython: language_level=3

from libc.stdint cimport uint32_t
from libc.stdlib cimport strtol
from libc.string cimport strchr
from libc.string cimport strlen
from libc.string cimport memset
import numpy as np
cimport numpy as cnp

import cython

@cython.boundscheck(False)
@cython.wraparound(False)
cdef uint32_t ip_to_int(char* ip):

    # Check if the input string is at least 7 characters long
    if strlen(ip) < 7:
        raise ValueError("Invalid IP address: too short")

    cdef uint32_t result = 0
    cdef uint32_t num = 0
    cdef int shift = 24  # Start with the leftmost byte
    cdef char* end

    # Convert each part of the IP to an integer
    for _ in range(4):
        num = strtol(ip, &end, 10)  # Convert substring to long
        if num > 255 or ip == end or (end[0] not in (b'.', b'\0') and _ < 3):  # Validate octet and check for non-digit characters
            raise ValueError("Invalid IP address: invalid part")
        result += num << shift
        shift -= 8
        if end[0] == b'\0':  # Check if end of string
            break
        ip = end + 1  # Move to the next part

    if shift != -8 or end[0] != b'\0':  # Ensure exactly 4 octets and end of string
        raise ValueError("Invalid IP address: not enough parts")

    return result

@cython.boundscheck(False)
@cython.wraparound(False)
def ip_in_cidr(cnp.ndarray ip_addresses, str cidr):

    # CIDR validation...
    if '/' not in cidr or not 0 <= int(cidr.split('/')[1]) <= 32:
        raise ValueError("Invalid CIDR notation")

    cdef uint32_t base_ip, netmask, ip_int
    cdef int mask_size
    cdef str base_ip_str
    cdef list cidr_parts = cidr.split('/')

    base_ip_str, mask_size = cidr_parts[0], int(cidr_parts[1])
    netmask = (0xFFFFFFFF << (32 - mask_size)) & 0xFFFFFFFF

    base_ip = ip_to_int(base_ip_str.encode('utf-8'))

    cdef unsigned char[:] result = np.zeros(ip_addresses.shape[0], dtype=np.bool_)

    for i in range(ip_addresses.shape[0]):
        ip_int = ip_to_int(ip_addresses[i].encode('utf-8'))
        result[i] = (ip_int & netmask) == base_ip

    return np.asarray(result, dtype=bool)
